import typing
import pathlib
import itertools
import collections
import dataclasses

from .tokenizer import TokenType, iter_tokens, get_name
from commonnexus.command import Command
from commonnexus.blocks import Block

__all__ = ['Config', 'Nexus']

NEXUS = '#NEXUS'


@dataclasses.dataclass
class Config:
    """
    The global behaviour of a :class:`Nexus` instance can be configured.
    The available configuration options are set and accessed from an instance of `Config`.
    """
    #: Specifies whether "-", aka ASCII hyphen-minus, is considered punctuation or not.
    hyphenminus_is_text: bool = True
    #: Specifies whether "*", aka asterisk, is considered punctuation or not.
    asterisk_is_text: bool = True
    #: Specifies whether Newick nodes for TREEs are constructed by parsing the Newick string or
    #: from the Nexus tokens. The latter is slightly faster but will bypass some input validation.
    validate_newick: bool = False
    #: Specifies whether unsupported NEXUS commands/options are ignored or raise an error. Note \
    #: that the effect of this option may only set in when a block or command is accessed.
    ignore_unsupported: bool = True
    #: Specifies the text encoding of a NEXUS file.
    encoding: str = 'utf8'
    #: The NEXUS spec does not explicitly state a default value for the MATCHCHAR directive in the
    #: FORMAT command of a CHARACTERS block. `commonnexus` - in agreement with many NEXUS files
    #: encountered "in the wild" - assumes a default of ".". To force no default value for
    #: MATCHCHAR, e.g. because matrix data uses "." as regular state symbol, set
    #: `no_default_matchchar` to `True`.
    no_default_matchchar: bool = False
    #: Sometimes the NEXUS spec is not followed entirely by files found in the wild. If somewhat
    #: lax interpretation does not lead to ambiguities, that's what commonnexus does. To force
    #: stricter adherence to the spec, set `strict` to `True`.
    strict: bool = False


class Nexus(list):
    """
    A NEXUS object implemented as list of tokens with methods to access newick constituents.

    From the spec:

        The tokens in a NEXUS file are organized into commands, which are in turn organized into
        blocks.

    This is reflected in the ``Nexus`` object. The ``Nexus`` object is just a ``list`` of
    :class:`Commands <Command>`, and has a property :meth:`Nexus.blocks` giving access to
    commands grouped by block:

    .. code-block::

        >>> nex = Nexus('#NEXUS BEGIN myblock; mycmd a b c; END;')
        >>> nex[0].__class__
        <class 'commonnexus.nexus.Command'>
        >>> len(nex.blocks['MYBLOCK'])
        1

    .. note::

        NEXUS is for the most part case-insensitive. `commonnexus` reflects this by giving all
        blocks and commands uppercase names. Thus, even if a command or block has a lowercase or
        mixed-case name in the file, the corresponding ``Command`` or ``Block`` object must be
        addressed using the uppercase name.
    """
    def __init__(self,
                 s: typing.Optional[typing.Union[typing.Iterable, typing.List[Command]]] = None,
                 block_implementations: typing.Optional[typing.Dict[str, Block]] = None,
                 config: typing.Optional[Config] = None,
                 **kw):
        """
        :param s: The NEXUS content.
        :param block_implementations: Custom implementations for non-public blocks.
        :param config: Configuration.
        :param kw: If no :class:`Config` object is passed as `config`, keyword parameters will be \
        interpreted as configuration options. Thus,

        .. code-block:: python

            >>> nex = Nexus(encoding='latin')

        is a shortcut for

        .. code-block:: python

            >>> nex = Nexus(config=Config(encoding='latin'))
        """
        self.cfg = config or Config(**kw)
        self.trailing_whitespace = []
        self.block_implementations = {}
        for cls in Block.__subclasses__():
            self.block_implementations[cls.__name__.upper()] = cls
            for scls in cls.__subclasses__():
                self.block_implementations.setdefault(scls.__name__.upper(), scls)

        self.block_implementations.update(block_implementations or {})
        s = s or NEXUS

        if isinstance(s, str):
            s = iter(s)

        if not isinstance(s, list):
            nexus, commands, tokens = False, [], []
            for token in itertools.dropwhile(
                    lambda t: t.type == TokenType.WHITESPACE, iter_tokens(s)):
                if not nexus:
                    assert token.type == TokenType.WORD and token.text.upper() == NEXUS, \
                        "No #NEXUS token found."
                    nexus = True
                else:
                    tokens.append(token)
                    if token.is_semicolon:
                        commands.append(Command(tuple(tokens)))
                        tokens = []
            self.trailing_whitespace = tokens
            s = commands
        list.__init__(self, s)

    @classmethod
    def from_file(cls,
                  p: typing.Union[str, pathlib.Path],
                  config: typing.Optional[Config] = None,
                  **kw) -> 'Nexus':
        """
        Instantiate a `Nexus` object from the contents of a NEXUS file.

        :param p: Path of the file.
        :param config: An optional configuration object.
        :param kw: Configuration options, if no `Config` object is passed in.
        :return: A `Nexus` instance.
        """
        config = config or Config(**kw)
        p = pathlib.Path(p)
        not_utf8 = False
        with p.open(encoding=config.encoding) as f:
            try:
                return cls(itertools.chain.from_iterable(f), config=config)
            except UnicodeDecodeError:
                not_utf8 = config.encoding == 'utf8'
        if not_utf8:
            # We don't want to do a lot of guessing, but if the default encoding was tried and
            # didn't work, we try with the old-time favourite "latin1":
            config.encoding = 'latin1'
            with p.open(encoding=config.encoding) as f:
                return cls(itertools.chain.from_iterable(f), config=config)

    @classmethod
    def from_blocks(cls, *blocks):
        res = cls()
        for block in blocks:
            res.append_block(block)
        return res

    @property
    def blocks(self) -> typing.Dict[str, typing.List[Block]]:
        """
        A `dict` mapping uppercase block names to lists of instances of these blocks ordered as they
        appear in the NEXUS content.

        For a shortcut to access blocks which are known to appear just once in the NEXUS content,
        see :meth:`Nexus.__getattribute__`.
        """
        res = collections.defaultdict(list)
        for block in self.iter_blocks():
            res[block.name].append(block)
        return res

    def __getattribute__(self, name):
        """
        NEXUS does not make any prescriptions regarding how many blocks with the same name may
        exist in a file. Thus, the primary way to access blocks is by looking up the list of blocks
        for a given name in :meth:`Nexus.blocks`. If it can be assumed that just one block for a
        name exists, or only the first block with that name is of interest, this block can also be
        accessed as `Nexus.<BLOCK_NAME>`, i.e. using the uppercase block name as attribute of the
        `Nexus` instance.

        .. code-block:: python

            >>> nex = Nexus('#NEXUS begin block; cmd; end;')
            >>> nex.BLOCK.name
            'BLOCK'
            >>> len(nex.BLOCK.commands)
            1
        """
        if name.isupper():
            try:
                return self.blocks[name][0]
            except IndexError:
                return None
        return list.__getattribute__(self, name)

    def __str__(self):
        """
        The string representation of a `Nexus` object is just its NEXUS content.

        .. code-block:: python

            >>> nex = Nexus()
            >>> nex.append_block(Block.from_commands([]))
            >>> print(nex)
            #NEXUS
            BEGIN BLOCK;
            END;
        """
        return NEXUS \
            + ''.join(''.join(str(t) for t in cmd) for cmd in self) \
            + ''.join(str(t) for t in self.trailing_whitespace)

    def to_file(self, p: typing.Union[str, pathlib.Path]):
        """
        Write the NEXUS content of a `Nexus` object to a file.
        """
        p = pathlib.Path(p)
        p.write_text(str(self), encoding=self.cfg.encoding)

    def iter_comments(self):
        for cmd in self:
            yield from (t for t in cmd if t.type == TokenType.COMMENT)

    @property
    def comments(self) -> typing.List[str]:
        """
        Comments may appear anywhere in a NEXUS file. Thus, they are the only kind of tokens not
        really grouped into a command.

        While comments **in** commands can also be accessed from the command, comments preceding
        any command (and all others) can accessed via this property.

        .. code-block:: python

            >>> nex = Nexus("#nexus [created by commonnexus] begin block; cmd [does nothing]; end;")
            >>> nex.BLOCK.CMD.comments
            ['does nothing']
            >>> nex.comments[0]
            'created by commonnexus'
        """
        return [t.text for t in self.iter_comments()]

    def iter_blocks(self):
        block = None
        for command in itertools.dropwhile(lambda c: not c.is_beginblock, self):
            if command.is_endblock:
                block.append(command)
                # Look up a suitable Block implementation.
                name = get_name(block[0].iter_payload_tokens())
                yield self.block_implementations.get(name, Block)(self, block)
                block = None
            elif command.is_beginblock:
                block = [command]
            elif block is not None:
                block.append(command)

    def validate(self, log=None):
        valid = True
        for block in self.iter_blocks():
            #
            # FIXME: we can do a lot of validation here! If block.__commands__ is a list, there is
            # some fixed order between commands.
            # If Payload.__multivalued__ == False, only one command instance is allowed, ...
            #
            valid = valid and block.validate(log=log)
        return valid

    def get_numbers(self, object_name, items):
        """
        Determine object numbers suitable for inclusion in a set spec.
        """
        if object_name == 'TAXON':
            return [str(i + 1) for i, tax in enumerate(self.taxa)
                    if (tax in items) or (str(i + 1) in items)]
        if object_name == 'CHARACTER':
            charlabels, statelabels = self.characters.get_charstatelabels()
            return [str(i) for i, label in charlabels.items()
                    if (label in items) or (str(i) in items)]
        if object_name == 'TREE':
            return [str(i + 1) for i, tree in enumerate(self.TREES.trees)
                    if (tree.name in items) or (str(i + 1) in items)]
        raise NotImplementedError(object_name)  # pragma: no cover

    def resolve_set_spec(self, object_name, spec, chars=None):
        """
        Resolve a set spec to a list of included items, specified by label or number.

        :param object_name:
        :param spec:
        :return:
        """
        def numbers(maxn):
            res = []
            start, r = None, False
            for token in spec:
                if not start:
                    start = token
                else:
                    if token == '-':
                        r = True
                    else:
                        if r:
                            res.extend(list(
                                range(int(start), int(token if token != '.' else maxn) + 1)))
                            r, start = False, None
                        else:
                            res.append(int(start))
                            start = token
            if start:
                res.append(int(start))
            return res

        object_name = object_name.upper()
        assert object_name in ['TAXON', 'CHARACTER', 'STATE', 'TREE']
        n, labels = None, None
        if object_name == 'TAXON':
            if self.TAXA:
                # FIXME: What if there's more than one TAXA block?
                n = self.TAXA.DIMENSIONS.ntax
                labels = self.TAXA.TAXLABELS.labels
            elif self.CHARACTERS and self.CHARACTERS.DIMENSIONS.newtaxa:
                n = self.CHARACTERS.DIMENSIONS.ntax
                if self.CHARACTERS.TAXLABELS:
                    labels = self.CHARACTERS.TAXLABELS.labels

        if object_name == 'CHARACTER':
            block = self.CHARACTERS or self.DATA
            if block:
                labels, _ = block.get_charstatelabels()
                n = len(labels)

        if object_name == 'STATE':
            chars = chars or []  # Labeled states make only sense in relation to characters.
            block = self.CHARACTERS or self.DATA
            if block:
                _, slabels = block.get_charstatelabels()
                res = collections.defaultdict(list)
                # need to transpose
                for char, states in slabels.items():
                    if char in chars:
                        res[char] = [
                            label or state for state, label in states.items() if state in spec]
                return res
            return  # pragma: no cover

        if object_name == 'TREE':
            labels = {i + 1: tree.name for i, tree in enumerate(self.TREES.trees)}

        if not labels:
            assert n
            labels = {i + 1: str(i + 1) for i in range(n)}
        return [labels[n] for n in numbers(n)]

    def remove_block(self, block: Block):
        for cmd in block:
            self.remove(cmd)

    def append_block(self, block: Block):
        self.extend(block)

    def prepend_block(self, block: Block):
        for cmd in reversed(block):
            self.insert(0, cmd)

    def replace_block(self,
                      old: Block,
                      new: typing.Union[Block, typing.List[typing.Tuple[str, str]]]):
        bname = old.name
        for i, cmd in enumerate(self):
            if cmd is old[0]:
                break
        else:
            raise ValueError('Block not found')  # pragma: no cover

        for cmd in old:
            self.remove(cmd)

        if isinstance(new, Block):
            new.nexus = self
            for cmd in reversed(new):
                self.insert(i, cmd)
        else:
            self.insert(i, Command.from_name_and_payload('END'))
            for n, payload in reversed(new):
                self.insert(i, Command.from_name_and_payload(n, payload))
            self.insert(i, Command.from_name_and_payload('BEGIN', bname))

    def append_command(self, block, name, payload=None):
        self.insert(
            self.index(block[-1]),
            Command.from_name_and_payload(name, payload))

    # Shortcuts:
    @property
    def characters(self) -> typing.Union[Block, None]:
        """
        Shortcut to get around the DATA/CHARACTERS ambiguity.

        I.e. if one is interested in the characters matrix of a NEXUS file no matter whether this
        is included in a DATA or CHARACTERS block, ``Nexus.characters.get_matrix()`` can be used
        rather than ``(Nexus.DATA or NEXUS.CHARACTERS).get_matrix()``.

        :return: The first DATA or CHARACTERS block.
        """
        assert not (self.DATA and self.CHARACTERS)
        return self.DATA or self.CHARACTERS

    @property
    def taxa(self) -> typing.Optional[typing.List[str]]:
        """
        Shortcut to retrieve the list of taxa a NEXUS file provides data on.

        :return: The list of taxa labels used in a NEXUS file.

        .. note::

            There are various ways to encode taxa labels in a NEXUS file. This method looks up
            different places ordered by explicitness, i.e.

            1. A TAXLABELS command in a TAXA block.
            2. A TAXLABELS command in a DATA or CHARACTERS block.
            3. Taxa labels given implicitly as labels in a MATRIX command.
            4. A TAXLABELS command in a DISTANCES block.
            5. Taxa labels given implicitly as labels in a DISTANCES.MATRIX command.
            6. Taxa labels given as mappings in the TRANSLATE command of a TREES block.
            7. Taxa labels given implicitly as node names in the Newick representation of a tree in
               a TREE command in a TREES block.

        .. warning::

            Taxa descriptions in NEXUS may be inconsistent, e.g. a NEXUS file might contain a
            TAXA block, but introduce new taxa via NEWTAXA/TAXLABELS in a CHARACTERS block.
            `commonnexus` does not make an effort to check for consistency.
        """
        if self.TAXA and len(self.blocks['TAXA']) == 1:
            return list(self.TAXA.TAXLABELS.labels.values())
        if self.characters:
            return list(self.characters.get_matrix())
        if self.DISTANCES:
            return list(self.DISTANCES.get_matrix())
        if self.TREES:
            if self.TREES.TRANSLATE:
                return list(self.TREES.TRANSLATE.mapping.values())
            return [node.name for node in self.TREES.TREE.newick.walk() if node.name]



from commonnexus.tools import normalise
from commonnexus.cli_util import add_nexus, add_flag


def help():
    return normalise.__doc__


def register(parser):
    add_nexus(parser)
    add_flag(parser, 'strip-comments', 'Remove non-command comments.')


def run(args):
    print(normalise.normalise(args.nexus, strip_comments=args.strip_comments))



import typing
import decimal
import warnings
import itertools
import collections

from commonnexus.tokenizer import iter_words_and_punctuation, iter_lines, Word
from .base import Block, Payload
from . import characters
from . import taxa

ODict = typing.OrderedDict


class Dimensions(characters.Dimensions):
    """
    The NTAX subcommand of this command is needed to process the matrix when some defined taxa are
    omitted from the distance matrix. The NCHAR subcommand is optional and can be used to indicate
    the number of characters for those analyses that need to know how many characters (if any) were
    used to calculate the distances. NCHAR is not required for successful reading of the matrix.
    As for the CHARACTERS and UNALIGNED block, taxa can be defined in a DISTANCES block if NEWTAXA
    precedes the NTAXA subcommand in the DIMENSIONS command. It is advised that new taxa not be
    defined in a DISTANCES block, for the reasons discussed in the description of the DATA block.
    NEWTAXA, if present, must be appear before the NTAX subcommand.

    :ivar bool newtaxa:
    :ivar typing.Optional[int] nchar:
    :ivar int ntax:
    """
    def check(self):
        assert (not self.newtaxa) or self.ntax


class Format(Payload):
    """
    This command specifies the formatting of the MATRIX. The [NO]LABELS and MISSING subcommands are
    as described in the CHARACTERS block.

    1. TRIANGLE = {LOWER | UPPER | BOTH}. This subcommand specifies whether only the lower left
       half of the matrix is present, or only the upper right, or both halves. Below is one example
       of an upper triangular matrix and one of a matrix with both halves included.

       .. code-block::

            BEGIN DISTANCES;
                FORMAT TRIANGLE=UPPER;
                MATRIX
                    taxon_1 0.0  1.0  2.0  4.0  7.0
                    taxon_2      0.0  3.0  5.0  8.0
                    taxon_3           0.0  6.0  9.0
                    taxon_4                0.0 10.0
                    taxon_5                     0.0;
            END;

       .. code-block::

            BEGIN DISTANCES;
                FORMAT TRIANGLE = BOTH;
                MATRIX
                    taxon_1  0    1.0  2.0  4.0  7.0
                    taxon_2  1.0  0    3.0  5.0  8.0
                    taxon_3  2.0  3.0  0    6.0  9.0
                    taxon_4  4.0  5.0  6.0  0   10.0
                    taxon_5  7.0  8.0  9.0 10.0  0;
            END;

    2. DIAGONAL. If DIAGONAL is turned off, the diagonal elements are not included:

       .. code-block::

            FORMAT NODIAGONAL;
            MATRIX
                taxon_1
                taxon_2  1.0
                taxon_3  2.0  3.0
                taxon_4  4.0  5.0  6.0
                taxon_5  7.0  8.0  9.0 10.0;

       If TRIANGLE is not BOTH and DIAGONAL is turned off, then there will be one row that contains
       only the name of a taxon. This row is required. If TRIANGLE=BOTH, then the diagonal must be
       included.
    3. INTERLEAVE. As in the CHARACTERS block, this subcommand indicates sections in the matrix,
       although interleaved matrices take a slightly different form for distance matrices:

       .. code-block::

            taxon_1  0
            taxon_2  1  0
            taxon_3  2  3  0
            taxon_4  4  5  6
            taxon_5  7  8  9
            taxon_6 11 12 13
            taxon_4  0
            taxon_5 10  0
            taxon_6 14 15  0;

       As in the CHARACTERS block, newline characters in interleaved matrices are significant, in
       that they indicate a switch to a new taxon.
    """
    def __init__(self, tokens, nexus=None):
        super().__init__(tokens, nexus=nexus)
        self.missing = '?'
        self.labels = True
        self.interleave = False
        self.diagonal = True
        self.triangle = 'LOWER'

        if tokens is None:
            return

        words = iter_words_and_punctuation(self._tokens, nexus=nexus)

        def word_after_equals():
            n = next(words)
            assert n.text == '='
            res = next(words)
            return res if isinstance(res, str) else res.text

        while 1:
            try:
                word = next(words)
                subcommand = None
                if isinstance(word, str):
                    subcommand = word.upper()
                if subcommand in ['TRIANGLE', 'MISSING']:
                    setattr(self, subcommand.lower(), word_after_equals())
                elif subcommand in ['INTERLEAVE']:
                    setattr(self, subcommand.lower(), True)
                elif subcommand in ['NOLABELS', 'LABELS', 'NODIAGONAL', 'DIAGONAL']:
                    setattr(self, subcommand.replace('NO', '').lower(), 'NO' not in subcommand)
            except StopIteration:
                break
        self.triangle = self.triangle.upper()


class Taxlabels(taxa.Taxlabels):
    """
    This command allows specification of the names and ordering of the taxa. It serves to define
    taxa and is allowed only if the NEWTAXA token is included in the DIMENSIONS statement.
    """


class Matrix(Payload):
    """
    This command contains the distance data.

    .. note::

        Since reading the matrix data only makes sense if information from other commands - in
        particular :class:`FORMAT <Format>` - is considered, the ``Matrix`` object does not have
        any attributes for data access. Instead, the matrix data can be read via
        :meth:`Distances.get_matrix`.
    """


class Distances(Block):
    """
    This block contains distance matrices. Taxa are usually not defined in a DISTANCES block; if
    they are not, this block must be preceded by a block that defines taxon labels and ordering
    (e.g., TAXA).
    The syntax of the block is as follows:

    .. rst-class:: nexus

        | BEGIN DISTANCES;
        |   [:class:`DIMENSIONS <Dimensions>` [NEWTAXA] NTAX=num-taxa NCHAR=num-characters;]
        |   [:class:`FORMAT <Format>`
        |     [TRIANGLE={LOWER | UPPER | BOTH}]
        |     [[NO]DIAGONAL]
        |     [[NO]LABELS]
        |     [MISSING=SYMBOL]
        |     [INTERLEAVE]
        |   ;]
        |   [:class:`TAXLABELS <Taxlabels>` taxon-name [taxon-name...];]
        |   :class:`MATRIX <Matrix>` distance-matrix;
        | END;

    Commands must appear in the order listed. Only one of each command is allowed per block.
    """
    __commands__ = [Dimensions, Format, Taxlabels, Matrix]

    def get_matrix(self) -> ODict[str, ODict[str, typing.Union[None, decimal.Decimal]]]:
        """
        :return: A full distance matrix encoded as nested ordered dictionaries.

        .. code-block:: python

            >>> from commonnexus import Nexus
            >>> nex = Nexus('''#NEXUS
            ... BEGIN DISTANCES;
            ...     DIMENSIONS NEWTAXA NTAX=5;
            ...     TAXLABELS taxon_1 taxon_2 taxon_3 taxon_4 taxon_5;
            ...     FORMAT TRIANGLE=UPPER;
            ...     MATRIX
            ...         taxon_1 0.0  1.0  2.0  4.0  7.0
            ...         taxon_2      0.0  3.0  5.0  8.0
            ...         taxon_3           0.0  6.0  9.0
            ...         taxon_4                0.0 10.0
            ...         taxon_5                     0.0;
            ... END;''')
            >>> nex.DISTANCES.get_matrix()['taxon_3']['taxon_1']
            Decimal('2.0')
        """
        format = self.FORMAT or Format(None)

        ntax, taxlabels = None, {}
        if self.DIMENSIONS:
            ntax = self.DIMENSIONS.ntax
        if self.TAXLABELS:
            taxlabels = self.TAXLABELS.labels
            ntax = self.DIMENSIONS.ntax
        elif self.nexus.TAXA:
            taxlabels = self.nexus.TAXA.TAXLABELS.labels
            ntax = self.nexus.TAXA.DIMENSIONS.ntax

        res = collections.OrderedDict()
        label, entries = None, []

        def required_cols(row_index=None):
            # The number of required entries for a distance matrix row depends on TRIANGLE,
            # DIAGONAL and the row index.
            if format.triangle == 'BOTH':  # Each row has entries for all taxa.
                ncols = ntax
            elif format.triangle == 'LOWER':  # Each row has one more entry than the previous one.
                if row_index:
                    ncols = row_index
                else:
                    ncols = 1 if not res else \
                        len(list(res.values())[-1]) + (2 if format.diagonal is False else 1)
            else:  # Each row has one entry less than the previous row.
                if row_index:
                    ncols = ntax - row_index + 1
                else:
                    ncols = ntax - len(list(res.values()))
            if not format.diagonal:
                # And if the diagonal is missing, we expect one entry less in all cases.
                ncols -= 1
            return ncols

        # Now read the matrix data:
        for i, line in enumerate(
            list(iter_lines(self.MATRIX._tokens)) if format.interleave else [self.MATRIX._tokens],
            start=1
        ):
            words = iter_words_and_punctuation(line, nexus=self.nexus)

            if (not format.labels) and required_cols() == 0 and 1 not in res:
                # NODIAGONAL NOLABELS TRIANGLE=LOWER
                res[1] = []

            while 1:
                try:
                    t = next(words)
                    if (format.labels is not False) and label is None:
                        assert isinstance(t, str)
                        label = t
                        if not format.diagonal and not format.interleave and \
                                format.triangle == 'LOWER' and not res:
                            # We're done with this row after the label.
                            res[label] = []
                            label = None
                        continue
                    entries.append(None if t == format.missing else decimal.Decimal(t))
                    if not format.interleave and (len(entries) == required_cols()):
                        res[label or (len(res) + 1)] = entries
                        label, entries = None, []
                except StopIteration:
                    break
            if format.interleave:
                # We collected a row of entries, now append them to the correct taxon.
                # If we have a label, that's easy.
                if not label:
                    for ri in taxlabels:
                        if ri not in res:
                            # First pass must go through all taxa!
                            label = ri
                            break
                    if not label:
                        # The next label to append entries to is the next one still in need of
                        # entries!
                        for row_index in res:
                            if len(res[row_index]) < required_cols(row_index):
                                label = row_index
                                break
                if label not in res:
                    res[label] = []
                res[label].extend(entries)
                label, entries = None, []

        if format.labels is False:
            assert taxlabels
        elif not taxlabels:
            taxlabels = collections.OrderedDict((i, label) for i, label in enumerate(res, start=1))

        # We pad the result rows with None columns on the left as necessary, to make lookup by
        # column index work.
        if format.triangle == 'UPPER':
            for i, key in enumerate(res):
                res[key] = [None] * (i if format.diagonal else i + 1) + res[key]
            if format.diagonal is False and format.labels is False:
                assert (ntax) not in res, str(res.keys())
                res[ntax] = [None]
        elif format.triangle == 'BOTH' and not format.diagonal:
            for i, key in enumerate(res):
                res[key].insert(i, None)

        # Match matrix rows to expected taxa:
        validtaxa = set(str(k) for k in taxlabels).union(taxlabels.values())
        restaxa = {str(k): k for k in res}
        if not set(restaxa).issubset(validtaxa):
            warnings.warn('Pruning undeclared taxa from DISTANCES matrix.')
            for taxon in set(restaxa) - validtaxa:
                del res[restaxa[taxon]]

        if len(res) < len(taxlabels):
            # Not all taxa appear in the matrix. Prune expected taxa to make lookup work.
            for k in list(taxlabels.keys()):
                if (str(k) not in restaxa) and (taxlabels[k] not in restaxa):
                    del taxlabels[k]
            taxlabels = collections.OrderedDict(
                (i, label) for i, label in enumerate(taxlabels.values(), start=1))

        res = {taxlabels.get(k, k): v for k, v in res.items()}
        assert set(res.keys()).issubset(taxlabels.values()), "Unmatched taxa in DISTANCES matrix."

        # Now populate a complete matrix with the data read from the tokens:
        matrix = collections.OrderedDict([
            (label, collections.OrderedDict([(ll, None) for ll in taxlabels.values()]))
            for label in taxlabels.values()])
        for (na, la), (nb, lb) in itertools.combinations_with_replacement(taxlabels.items(), r=2):
            if na == nb and format.diagonal is False:
                matrix[la][lb] = 0
                continue
            if format.triangle == 'BOTH':
                matrix[la][lb] = res[la][nb - 1]
                matrix[lb][la] = res[lb][na - 1]
            else:
                val = res[la][nb - 1] if format.triangle == 'UPPER' else res[lb][na - 1]
                if nb != na:
                    matrix[la][lb] = matrix[lb][la] = val
                else:
                    matrix[la][lb] = val

        return matrix

    @classmethod
    def from_data(cls,
                  matrix: typing.OrderedDict[str, typing.OrderedDict[
                      str, typing.Union[None, float, int, decimal.Decimal]]],
                  taxlabels: bool = False,
                  TITLE: typing.Optional[str] = None,
                  ID: typing.Optional[str] = None,
                  LINK: typing.Optional[typing.Union[str, typing.Tuple[str, str]]] = None,
                  **kw) -> 'Block':
        """
        Create a DISTANCES block from the distance matrix `matrix`.

        :param matrix: The distance matrix as dict mapping taxon labels to dicts mapping taxon \
        labels to numbers. A "full" matrix is expected here, just like it is returned from \
        :meth:`Distances.get_matrix`.
        :param taxlabels: Whether to include a TAXLABELS command.
        """
        nexus = kw.pop('nexus', None)

        dimensions = 'NTAX={}'.format(len(matrix))
        if taxlabels:
            dimensions = 'NEWTAXA NTAX={} {}'.format(len(matrix), dimensions)
        cmds = [
            # DIMENSIONS: necessary if not all taxa are present in the matrix!
            ('DIMENSIONS', dimensions),
            ('FORMAT', 'TRIANGLE=BOTH MISSING=?'),
        ]
        maxlen, tlabels = 0, {}
        for taxon in matrix:  # We compute maximum taxon label length for pretty printing.
            tlabels[taxon] = Word(taxon).as_nexus_string()
            maxlen = max([maxlen, len(tlabels[taxon])])

        if taxlabels:
            cmds.append(('TAXLABELS', ' '.join(tlabels.values())))

        mrows = []
        for taxon, dists in matrix.items():
            row = [tlabels[taxon].ljust(maxlen)]
            row.extend(['?' if v is None else str(v) for v in dists.values()])
            mrows.append(' '.join(row))
        cmds.append(('MATRIX', ''.join('\n' + row for row in mrows) + '\n'))
        return cls.from_commands(cmds, nexus=nexus, TITLE=TITLE, LINK=LINK, ID=ID)



"""
Tools to manipulate matrices as returned by
:meth:`commonnexus.blocks.characters.Characters.get_matrix`.
"""
import string
import textwrap
import collections
import typing

from commonnexus.blocks.characters import GAP, State, StateMatrix

HashableState = typing.Union[None, str, typing.FrozenSet[str], typing.Tuple[str]]


class CharacterMatrix(collections.OrderedDict):
    """
    A wrapper for the nested ordered dicts returned by
    :meth:`commonnexus.blocks.characters.Characters.get_matrix`, providing
    simpler access to some properties of the data and some conversion functionality.
    """
    def iter_rows(self) -> typing.Generator[typing.List[State], None, None]:
        """Iterate lists of states per taxon."""
        for row in self.values():
            yield list(row.values())

    def iter_columns(self) -> typing.Generator[typing.List[State], None, None]:
        """Iterate lists of states per character."""
        for char in self.characters:
            yield [self[taxon][char] for taxon in self.taxa]

    @property
    def taxa(self) -> typing.List[str]:
        """The list of taxa (labels or numbers) in a matrix."""
        return list(self.keys())

    @property
    def characters(self) -> typing.List[str]:
        """The list of characters (labels or numbers) in a matrix."""
        for row in self.values():
            return list(row.keys())

    @property
    def distinct_states(self) -> typing.Set[HashableState]:
        """
        The set of distinct states in a matrix (including missing and gap, if found).
        """
        res = set()
        for r in self.values():
            for v in r.values():
                res.add(frozenset(v) if isinstance(v, set) else v)
        return res

    @property
    def has_missing(self) -> bool:
        return None in self.distinct_states

    @property
    def has_gaps(self) -> bool:
        return GAP in self.distinct_states

    @property
    def has_uncertain(self) -> bool:
        for s in self.distinct_states:
            if isinstance(s, frozenset):
                return True
        return False

    @property
    def has_polymorphic(self) -> bool:
        for s in self.distinct_states:
            if isinstance(s, tuple):
                return True
        return False

    @property
    def symbols(self) -> typing.Set[typing.Union[str, typing.FrozenSet[str], typing.Union[str]]]:
        """The set of state symbols, excluding missing and gapped."""
        res = set()
        for s in self.distinct_states:
            if (s is not None) and (s != GAP):
                res |= set(s)
        return res

    @property
    def is_binary(self) -> bool:
        return self.symbols.issubset({'0', '1'}) and not self.has_gaps

    @classmethod
    def binarised(cls,
                  matrix: StateMatrix,
                  statelabels: typing.Optional[typing.Dict[str, typing.Dict[str, str]]] = None) \
            -> 'CharacterMatrix':
        statelabels = statelabels or {}
        matrix = cls(matrix)
        charstates = collections.defaultdict(set)
        for i, col in enumerate(matrix.iter_columns()):
            for v in col:
                if v is not None and v != GAP:
                    charstates[matrix.characters[i]] |= set(v)
        charstates = {k: sorted(v, key=lambda vv: str(vv)) for k, v in charstates.items()}
        new = collections.OrderedDict()

        for taxon, row in matrix.items():
            new[taxon] = collections.OrderedDict()
            for char, value in row.items():
                #
                # FIXME: don't binarise what's already binary!
                #
                for i, state in enumerate(charstates[char], start=1):
                    if value is None:
                        v = None
                    elif value == GAP:
                        v = GAP
                    else:
                        v = ({'1'} if isinstance(value, set) else '1') if state in value else '0'
                    statelabel = statelabels.get(char, {}).get(state) or state
                    new[taxon]['{}_{}'.format(char, statelabel)] = v
        return cls(new)

    @classmethod
    def multistatised(cls, matrix: StateMatrix, multicharlabel: typing.Optional[str] = None)\
            -> 'CharacterMatrix':
        """
        Convert character data of the form 0010000 to a single multi-state character.
        This kind of data may be obtained from coding wordlist data as "word belongs to cognate set"
        vectors.

        If 26..52 characters are given, RESPECTCASE is added to FORMAT, and A-Za-z is used as symbol
        set.
        """
        matrix = cls(matrix)
        available_states = list(string.ascii_uppercase + string.ascii_lowercase)
        assert matrix.is_binary, "All state symbols must be 0, 1 or None (missing)"
        assert len(matrix.characters) <= len(available_states), \
            "Too many characters to multistatise"

        multicharlabel = multicharlabel or '1'
        # Seed the resulting matrix with `None`, i.e. "missing" values.
        multistate_matrix = collections.OrderedDict(
            [(t, collections.OrderedDict([(multicharlabel, None)])) for t in matrix])
        for i, charlabel in enumerate(matrix.characters):
            for taxon in matrix:
                if matrix[taxon][charlabel] == '1':
                    if multistate_matrix[taxon][multicharlabel]:
                        multistate_matrix[taxon][multicharlabel] = tuple(
                            list(multistate_matrix[taxon][multicharlabel]) + [available_states[i]])
                    else:
                        multistate_matrix[taxon][multicharlabel] = available_states[i]
        return cls(multistate_matrix)

    @classmethod
    def from_characters(cls,
                        matrix: StateMatrix,
                        drop_chars: typing.Optional[typing.Iterable[str]] = None,
                        inverse: bool = False,
                        drop_uncertain: bool = False,
                        drop_polymorphic: bool = False,
                        drop_missing: bool = False,
                        drop_gapped: bool = False,
                        drop_constant: bool = False) -> 'CharacterMatrix':
        """
        :param chars:
        :param inverse:
        :return: A **new** matrix constructed as copy, omitting specified characters.
        """
        drop_chars = drop_chars or set()
        matrix = cls(matrix)
        taxa, characters = matrix.taxa, matrix.characters
        res = collections.OrderedDict([(t, collections.OrderedDict()) for t in matrix])
        for i, col in enumerate(matrix.iter_columns()):
            char = characters[i]
            if drop_chars and not inverse and (char in drop_chars):
                continue
            if drop_chars and inverse and (char not in drop_chars):
                continue
            if drop_uncertain and any(isinstance(v, set) for v in col):
                continue
            if drop_polymorphic and any(isinstance(v, tuple) for v in col):
                continue
            if drop_missing and any(v is None for v in col):
                continue
            if drop_gapped and any(v == GAP for v in col):
                continue
            if drop_constant and \
                    len(set(frozenset(v) if isinstance(v, set) else v for v in col)) == 1:
                continue
            for j, v in enumerate(col):
                res[taxa[j]][char] = v
        return cls(res)

    def to_phylip(self) -> str:
        def phylip_name(s):
            res = ''
            for c in s:
                if c in string.printable and (c not in '()[]:;,'):
                    res += c
                if len(res) >= 10:
                    break
            return res.ljust(10)

        if self.has_uncertain or self.has_polymorphic:
            raise ValueError('Cannot convert matrix with uncertain or polymorphic states.')
        if self.has_missing and '?' in self.symbols:
            raise ValueError('Missing symbol ? used as state symbol')  # pragma: no cover
        if self.has_gaps and '-' in self.symbols:
            raise ValueError('Gap symbol - used as state symbol')  # pragma: no cover

        res = ["    {}   {}".format(len(self.taxa), len(self.characters))]
        for taxon, states in self.items():
            seq = ''
            for state in states.values():
                assert isinstance(state, str) or state is None
                seq += '?' if state is None else ('-' if state == GAP else state)
            res.append('{}{}'.format(phylip_name(taxon), seq))
        return '\n'.join(res)

    def to_fasta(self) -> str:
        """
        :return: The character matrix serialized in the \
        `FASTA format <https://en.wikipedia.org/wiki/FASTA_format>`_
        """
        # convert states codes as digits to letters
        # convert missing *and* gap to '-'
        if self.has_uncertain or self.has_polymorphic:
            raise ValueError('Cannot convert matrix with uncertain or polymorphic states.')

        digits = {s: None for s in self.symbols if s in string.digits}
        if digits:
            if len(self.symbols) > len(string.ascii_uppercase):  # pragma: no cover
                raise ValueError('Too many symbols in matrix to replace digits with letters')
            for digit in sorted(digits):
                for c in string.ascii_uppercase:
                    if c not in self.symbols and (c not in digits.values()):
                        digits[digit] = c
                        break

        res = []
        for taxon, states in self.items():
            seq = ''.join(
                '-' if state is None or state == GAP else digits.get(state, state)
                for state in states.values())
            res.append('> {}'.format(taxon))
            res.extend(textwrap.wrap(seq, 70))
        return '\n'.join(res)

    @classmethod
    def from_fasta(cls, fasta: str) -> 'CharacterMatrix':
        """
        .. code-block:: python

            >>> from commonnexus import Nexus
            >>> from commonnexus.blocks import Data
            >>> from commonnexus.tools.matrix import CharacterMatrix
            >>> print(Nexus.from_blocks(Data.from_data(CharacterMatrix.from_fasta(
            ...     '> t1\\nABA BAA\\n> t2\\nBAB ABA'))))
            #NEXUS
            BEGIN DATA;
                DIMENSIONS NCHAR=6;
                FORMAT DATATYPE=STANDARD MISSING=? GAP=- SYMBOLS="AB";
                MATRIX
                t1 ABABAA
                t2 BABABA;
            END;
        """
        def get_row(nchar, seq, taxon):
            if nchar is not None:
                assert len(seq) == nchar, "Only aligned sequences can be converted."
            assert taxon is not None
            return len(seq), collections.OrderedDict([(str(i + 1), c) for i, c in enumerate(seq)])

        res = collections.OrderedDict()
        taxon, seq, nchar = None, [], None
        for line in fasta.split('\n'):
            if line.startswith('>'):
                if seq:
                    nchar, res[taxon] = get_row(nchar, seq, taxon)
                taxon, seq = line[1:].strip(), []
                continue
            for chunk in line.strip().split():
                for c in chunk:
                    if c not in string.digits:
                        seq.append(c)
        if seq:
            _, res[taxon] = get_row(nchar, seq, taxon)
        return cls(res)



from .base import Block


class Assumptions(Block):
    """
    .. warning::

        `commonnexus` doesn't provide any functionality - other than parsing as generic commands -
        for ``ASSUMPTIONS`` blocks.

        .. code-block:: python

            >>> nex = Nexus('''#NEXUS
            ...     BEGIN ASSUMPTIONS;
            ...         OPTIONS DEFTYPE=ORD;
            ...         USERTYPE my0rd = 4
            ...                 0 1 2 3
            ...                 . 1 2 3
            ...                 1 . 1 2
            ...                 2 1 . 1
            ...                 3 2 1 .;
            ...         USERTYPE myTree (CSTREE) = ((0,1) a,(2,3)b)c;
            ...         TYPESET * mixed=lRREv: 1 3 10, UNORD 5-7;
            ...         WTSET * one = 2 : 1-3 6 11-15, 3: 7 8;
            ...         WTSET two = 2:4 9, 3: 1-3 5;
            ...         EXSET nolarval = 1-9;
            ...         ANCSTATES mixed = 0: 1 3 5-8 11, 1: 2 4 9-15;
            ...     END;''')
            >>> str(nex.ASSUMPTIONS.ANCSTATES)
            'mixed = 0: 1 3 5-8 11, 1: 2 4 9-15'

    The ASSUMPTIONS block houses assumptions about the data or gives general directions as to how
    to treat them (e.g., which characters are to be excluded from consideration). The commands
    currently placed in this block were primarily designed for parsimony analysis. More commands,
    embodying assumptions useful in distance, maximum likelihood, and other sorts of analyses, will
    be developed in the future.
    For example, matrices specifying relative rates of character state change, useful for both
    distance and likelihood analyses, will eventually be included here.
    The general structure of the assumptions block is

    .. rst-class:: nexus

        | BEGIN ASSUMPTIONS;
        |   [OPTIONS [DEFTYPE = type-name]
        |     [POLYTCOUNT= {MINSTEPS | MAXSTEPS}]
        |     [GAPMODE= {MISSING | NEWSTATE}];]
        |     [USERTYPE type-name [ ( {STEPMATRIX | CSTREE} ) ] =UsERTYPE-description; ]
        |   [TYPESET [*] typeset-name [ ({STANDARD | VECTOR} ) ] = TYPESET-definition;]
        |   [WTSET [*] wtset-name [({STANDARD | VECTOR} {TOKENS | NOTOKENS})] = WrSET-defini tion;]
        |   [EXSET [*] exset-name [( {STANDARD | VECTOR} ) ] = character-set; ]
        |   [ANCSTATES [*] ancstates-name [({STANDARD | VECTOR} {[NO]TOKENS})] =
        |     ANCSTATES-definition;]
        | END;

    An example ASSUMPTIONS block follows:

    .. code-block::

        BEGIN ASSUMPTIONS;
            OPTIONS DEFTYPE=ORD;
            USERTYPE my0rd = 4
                0 1 2 3
                . 1 2 3
                1 . 1 2
                2 1 . 1
                3 2 1 .;
            USERTYPE myTree (CSTREE) = ((0,1) a,(2,3)b)c;
            TYPESET * mixed=lRREv: 1 3 10, UNORD 5-7;
            WTSET * one = 2 : 1-3 6 11-15, 3: 7 8;
            WTSET two = 2:4 9, 3: 1-3 5;
            EXSET nolarval = 1-9;
            ANCSTATES mixed = 0: 1 3 5-8 11, 1: 2 4 9-15;
        END;

    **USERTYPES** must be defined before they are referred to in any TYPESET.

    In earlier versions of MacClade and PAUP, TAXSET and CHARSET also appeared in the
    ASSUMPTIONS block. These now appear in the SETS block. There are a number of other commands
    in the ASSUMPTIONS block that also have SET in their name (e.g., WTSET, EXSET), but these
    commands assign values to objects, they do not define sets of objects, and therefore they
    do not belong in the SETS block. (Commands such as WTSET and EXSET are so named for
    historical reasons; although they might ideally be renamed CHARACTERWEIGHTS and
    EXCLUDEDCHARACTERS, doing so would cause existing programs to be incompatible with the file
    format.) We recommend that programs also accept TAXSET and CHARSET commands in the
    ASSUMPTIONS block so that older files can be read. In addition, the GAPMODE subcommand of
    the OPTIONS command of this block was originally housed in an OPTIONS command in the DATA
    block. Because this subcommand dictates how data are to be treated rather than providing
    details about the data themselves, it was moved into the ASSUMPTIONS block.

    **OPTIONS** — This command houses a number of disparate subcommands. They are all of the form
    subcommand=option.

    1. **DEFTYPE**. This subcommand specifies the default character type for parsimony analyses.
       Whenever a character's type is not explicitly stated, its type is taken to be the default
       type. Default DEFTYPE is UNORD (see the Appendix, character trans-
       formation type, for a definition of UNORD).
    2. **POLYTCOUNT**. Setting POLYTCOUNT to MINSTEPS specifies that trees with polytomies are to
       be counted (in parsimony analyses) in such a way that the number of steps for each character
       is the minimum number of steps for that character over any resolution of the polytomy. A
       tree length that is the sum of these minimum numbers of steps may be below the tree length
       of the most-parsimonious dichotomous resolution. Setting POLYTCOUNT to MAXSTEPS specifies
       that trees with polytomies are to be counted in such a way that occurrence of derived states
       on elements of a polytomy are to be counted as independent derivations. Such a tree length
       may be above the tree length of any fully dichotomous resolution. The NEXUS format does not
       specify a default value for POLYTCOUNT; the default value may differ from program to program.
    3. **GAPMODE**. This subcommand specifies how gaps are to be treated. GAPMODE=MISSING specifies
       that gaps are to be treated in the same way as missing data; GAPMODE=NEWSTATE specifies that
       gaps are to be treated as an additional state (for DNA/RNA/NUCLEOTIDE data, as a fifth base).

    **USERTYPE** —This command defines a character transformation type, as used in parsimony
    analysis to designate the cost of changes between states. There are several predefined
    character types (see character transformation type in the Appendix); USERTYPE allows additional
    character types to be created. USERTYPE is an object definition command with the exception that
    an asterisk cannot be used to indicate the default type (default type is stated in the OPTIONS
    command of the ASSUMPTIONS block). The standard defines no limit to the length of the type name,
    although individual programs might impose restrictions.

    STEPMATRIX format is

    .. code-block::

        USERTYPE myMatrix (STEPMATRIX) = n
            s s s s
            . k k k
            k . k k
            k k . k
            k k k .;

    where n is the number of rows and columns in the step matrix, the s's are state symbols, and
    the k's are the cost for going between states. The n can take any value >2. Diagonal elements
    may be listed as periods. If a change is to be prohibited, then one enters an "i" for infinity.
    Typically, the state symbols will be in sequence, but they need not be. The following matrices
    assign values identically:

    .. code-block::

        USERTYPE myMatrix (STEPMATRIX) =4
            0 1 2 3
            . 1 5 1
            1 . 5 1
            5 5 . 5
            1 1 5 . ;
        USERTYPE myMatrix2 (STEPMATRIX) =4
            2 0 3 1
            . 5 5 5
            5 . 1 1
            5 1 . 1
            5 1 1 .;

    The number of steps may be either integers or real numbers. The range of possible values will
    differ from program to program. Versions 3.0-3.04 of MacClade use the format name REALMATRIX
    rather than STEPMATRIX if the matrix contains real numbers. Future programs should treat
    REALMATRIX as a synonym of STEPMATRIX.

    CSTREE format is very similar to the TREE format in a TREES block. That is, character state
    trees are described in the parenthesis notation following the rules given for TREES of taxa.
    Instead of taxon labels, character state symbols are used. Thus,

    .. code-block::

        USERTYPE cstree-name (CSTREE) = [{list-of-subtrees)] [state-symbol]];

    where each subtree has the same format as the overall tree and the subtrees are separated by
    commas.

    **TYPESET** — This command specifies the type assigned to each character as used in parsimony
    analysis. This is a standard object definition command. Any characters not listed in the
    character-set have the default character type. The type names to be used are either the
    predefined ones or those defined in a USERTYPE command. Each value in a definition in VECTOR
    format must be separated by whitespace. The following are equivalent type sets:

    .. code-block::

        TYPESET mytypes = ORD: 1 4 6 , UNORD: 2 3 5 ;
        TYPESET mytypes (VECTOR) = ORD UNORD UNORD ORD UNORD ORD;

    **WTSET** — This command specifies the weights of each character. This is a standard object
    definition command. Any characters not listed in the character-set have weight 1. The weights
    may be either integers or real numbers. The minimum and maximum weight value will differ from
    program to program. Each value in a definition in VECTOR format must be separated by whitespace
    unless the NOTOKENS option is invoked, in which case no whitespace is needed and all weights
    must be integers in the range 0-9. The following are equivalent weight sets:

    .. code-block::

        WTSET mywts = 3 : 1 4 6, 1: 2 3 5;
        WTSET mywts (VECTOR) =3 1 1 3 1 3 ;

    In earlier versions of MacClade, the formatting subcommand REAL was used to indicate that
    real-valued weights were included in the WTSET. This subcommand is no longer in use; programs
    are expected to detect the presence of integral or real-value weights while reading the WTSET
    command.

    **EXSET** — This command specifies which characters are to be excluded from consideration. This
    is a standard object definition command. Any characters not listed in the character-set are
    included. The VECTOR format consists of 0's and l's: a 1 indicates that the character is to be
    excluded; whitespace is not necessary between 0's and l's.
    The following commands are equivalent and serve to exclude characters 5, 6, 7, 8, and 12.

    .. code-block::

        EXSET * toExclude = 5-8 12;
        EXSET * toExclude (VECTOR) = 000011110001;

    **ANCSTATES** — This command allows specification of ancestral states. This is a standard object
    definition command. Any valid state symbol can be used in the description for discrete data, and
    any valid value can be used for continuous data. TOKENS is the default for DATATYPE=CONTINUOUS;
    NOTOKENS is the default for all other DATATYPES. TOKENS is not allowed for DATATYPES DNA, RNA,
    and NUCLEOTIDE. If TOKENS is invoked, the standard three-letter amino acid abbreviations can be
    used with DATATYPE=PROTEIN and defined state names can be used for DATATYPE=STANDARD. NOTOKENS
    is not allowed for DATATYPE=CONTINUOUS. The following commands are equivalent:

    .. code-block::

        ANCSTATES a n c e s t o r = 0 :1-3 5-7 12, 1:4 8-10, 2 : 1 1 ;
        ANCSTATES a n c e s t o r (VECTOR) = 000100011120;
    """
    pass



"""
Manipulate the list of TAXA used in a NEXUS file.
"""
import collections

from commonnexus.cli_util import add_nexus, add_flag, ParserError, add_rename
from commonnexus.blocks import Data, Characters, Trees, Taxa, Distances
from commonnexus.blocks.characters import GAP


def register(parser):
    add_nexus(parser)
    parser.add_argument(
        '--drop',
        help="Comma-separated list of taxon labels or numbers to remove from the NEXUS. This will "
             "remove the taxa from the TAXA block, the relevant rows from a CHARACTERS (or DATA) "
             "matrix, and prune the specified taxa from any TREE in a TREES block.",
        type=lambda s: s.split(','),
        default=[])
    add_rename(parser, 'taxon')
    parser.add_argument(
        '--describe',
        help="Describe a named taxon, i.e. aggregate the data for the taxon in a NEXUS file.",
        default=None)
    add_flag(parser, 'check', 'Check whether taxa labels in a NEXUS file are used consistently.')


def run(args):
    taxa = collections.OrderedDict(
        [(str(i), taxon) for i, taxon in enumerate(args.nexus.taxa, start=1)])

    if args.drop or args.rename:
        if args.drop:
            to_drop, to_keep = collections.OrderedDict(), []
            for k, v in taxa.items():
                if k in args.drop or (v in args.drop):
                    to_drop[k] = v
                else:
                    to_keep.append(v)

            if args.nexus.characters:
                matrix = args.nexus.characters.get_matrix()
                for taxon in list(matrix.keys()):
                    if taxon in to_drop or (taxon in to_drop.values()):
                        del matrix[taxon]
                characters = (Data if args.nexus.characters.name == 'DATA' else Characters)\
                    .from_data(matrix)
                args.nexus.replace_block(args.nexus.characters, characters)
            if args.nexus.TREES:
                trees = []
                for tree in args.nexus.TREES.trees:
                    if args.nexus.TREES.TRANSLATE:
                        nwk = args.nexus.TREES.translate(tree)
                    else:
                        nwk = tree.newick
                    nwk.prune_by_names(list(to_drop) + list(to_drop.values()))
                    trees.append((tree.name, nwk, tree.rooted))
                args.nexus.replace_block(args.nexus.TREES, Trees.from_data(*trees))

            if args.nexus.DISTANCES:
                matrix = args.nexus.DISTANCES.get_matrix()
                for taxon in list(matrix.keys()):
                    if taxon in to_drop or (taxon in to_drop.values()):
                        del matrix[taxon]  # pragma: no cover

                for dists in matrix.values():
                    for taxon in list(dists.keys()):
                        if taxon in to_drop or (taxon in to_drop.values()):
                            del dists[taxon]  # pragma: no cover

                args.nexus.replace_block(args.nexus.DISTANCES, Distances.from_data(matrix))

            if args.nexus.NOTES:  # pragma: no cover
                args.log.warning('Changing taxon sets in NOTES block is not supported.')

            if args.nexus.TAXA:
                args.nexus.replace_block(args.nexus.TAXA, Taxa.from_data(to_keep))
            else:
                args.nexus.append_block(Taxa.from_data(to_keep))

        if args.rename:
            mapping = {}
            if callable(args.rename):
                for number, label in taxa.items():
                    new = args.rename(label)
                    if new != label:
                        taxa[number] = mapping[number] = mapping[label] = new
            else:
                old, new = args.rename
                for number, label in taxa.items():
                    if old == number or old == label:
                        taxa[number] = mapping[number] = mapping[old] = new
                        break
                else:
                    raise ParserError('No taxon matching {} found'.format(old))  # pragma: no cover

            if args.nexus.characters:
                matrix = args.nexus.characters.get_matrix()
                if set(matrix).intersection(mapping):
                    matrix = collections.OrderedDict(
                        [(mapping.get(k, k), v) for k, v in matrix.items()])
                    args.nexus.replace_block(args.nexus.characters, Characters.from_data(matrix))
            if args.nexus.DISTANCES:
                matrix = collections.OrderedDict()
                for tax, dists in args.nexus.DISTANCES.get_matrix().items():
                    matrix[mapping.get(tax, tax)] = collections.OrderedDict([
                        (mapping.get(k, k), v) for k, v in dists.items()])
                args.nexus.replace_block(args.nexus.DISTANCES, Distances.from_data(matrix))
            if args.nexus.TREES:
                labels = {}
                if args.nexus.TREES.TRANSLATE:
                    labels = collections.OrderedDict([
                        (k, mapping.get(v, v))
                        for k, v in args.nexus.TREES.TRANSLATE.mapping.items()])
                trees = []
                for tree in args.nexus.TREES.trees:
                    trees.append((tree.name, tree.newick.rename(**mapping), tree.rooted))
                args.nexus.replace_block(args.nexus.TREES, Trees.from_data(*trees, **labels))

            if args.nexus.NOTES:  # pragma: no cover
                args.log.warning('Changing taxon sets in NOTES block is not supported.')

            if args.nexus.TAXA:
                args.nexus.replace_block(args.nexus.TAXA, Taxa.from_data(taxa.values()))
            else:
                args.nexus.append_block(Taxa.from_data(taxa.values()))

        args.nexus.validate(args.log)
        print(args.nexus)
        return

    if args.check:
        if args.nexus.characters:
            mtaxa = list(args.nexus.characters.get_matrix())
            if len(set(mtaxa).intersection(set(taxa).union(taxa.values()))) < len(taxa):
                args.log.warning('Not all taxa appear in characters matrix')
        if args.nexus.DISTANCES:
            args.nexus.DISTANCES.get_matrix()
        if args.nexus.TREES:
            if args.nexus.TREES.TRANSLATE:
                if not set(args.nexus.TREES.TRANSLATE.mapping.values())\
                        .issubset(set(taxa.values())):
                    args.log.error('Invalid taxa labels in TREES TRANSLATE.')
            for tree in args.nexus.TREES.trees:
                nwk = args.nexus.TREES.translate(tree)
                leafnames = {n.name for n in nwk.walk() if n.name and n.is_leaf}
                if not leafnames.issubset(taxa.values()):
                    args.log.error('Invalid taxa labels as leaf name in TREE.')
                names = {n.name for n in nwk.walk() if n.name and not n.is_leaf}
                if not names.issubset(taxa.values()):
                    args.log.warning('Invalid taxa labels as inner node name in TREE.')
        return

    if args.describe:
        for num, name in taxa.items():
            if args.describe in (num, name):
                break
        else:
            raise ValueError('Taxon not found')  # pragma: no cover
        print('# {}\n'.format(name))
        charnotes = collections.defaultdict(list)
        if args.nexus.NOTES:
            for text in args.nexus.NOTES.texts:
                for char in text.characters or []:
                    charnotes[char].append(text.text)
        if args.nexus.characters:
            m = args.nexus.characters.get_matrix(labeled_states=True)
            print('Character | State | Notes')
            print('--- | --- | ---')
            for char, state in m[name].items():
                notes = '/'.join(charnotes.get(char, []))
                print('{} | {} | {}'.format(
                    char, '?' if state is None else ('-' if state == GAP else state), notes))
        if args.nexus.TREES:
            print('\n## Tree {}\n'.format(args.nexus.TREES.TREE.name))
            tree = args.nexus.TREES.translate(args.nexus.TREES.TREE)
            print('```')
            print(tree.ascii_art())
            print('```')
        if args.nexus.NOTES:
            for i, text in enumerate(args.nexus.NOTES.get_texts(taxon=name)):
                if i == 0:
                    print('\n## Notes\n')
                    print('Character | Note')
                    print('--- | ---')
                print('{} | {}'.format('/'.join(text.characters or []), text.text.strip()))
        return

    args.log.info('Taxa:\nnumber\tlabel')
    for i, taxon in taxa.items():
        print('{}\t{}'.format(i, taxon))



import typing
import warnings
import functools
import collections

import newick

from .base import Payload, Block
from commonnexus.util import log_or_raise
from commonnexus.tokenizer import TokenType, iter_words_and_punctuation, Word, Token

TreeSpec = typing.Tuple[str, typing.Union[str, newick.Node], typing.Union[None, bool]]


class Translate(Payload):
    """
    The tree description requires references to the taxa defined in a TAXA, DATA,
    CHARACTERS, UNALIGNED, or DISTANCES block. These references can be made using the label assigned
    to them in the TAXA or DATA blocks, their numbers, or a token specified in the TRANSLATE
    command. The TRANSLATE statement maps arbitrary labels in the tree specification to valid taxon
    names. If the arbitrary labels are integers, they are mapped onto the valid taxon names as
    dictated by the TRANSLATE command without any consideration of the order of the taxa in the
    matrix. Thus, if an integer is encountered in the tree description, a program first checks to
    see if it matches one of the arbitrary labels defined in the TRANSLATE command; only if
    no matching label is found will the integer be presumed to refer to the taxon in that
    position in the matrix (e.g., if the label in the description is 15, but this is not a label
    defined in the TRANSLATE command, a program should take this to refer to the 15th taxon).

    In the following example,

    .. code-block::

        BEGIN TAXA;
            TAXLABELS Scarabaeus Drosophila Aranaeus;
        END;
        BEGIN TREES;
            TRANSLATE beetle Scarabaeus, fly Drosophila, spider Aranaeus;
            TREE tree1 = ((1,2),3);
            TREE tree2 = ((beetle,fly),spider);
            TREE tree3= ((Scarabaeus,Drosophila),Aranaeus);
        END;

    the TRANSLATE command specifies that the label "beetle" can be used in the tree description to
    refer to Scarabaeus, "fly" to Drosophila, and "spider" to Aranaeus. This means that Scarabaeus
    can be referred to in a tree description as 1, Scarabaeus, or beetle. Thus, the three trees are
    identical.

    :ivar typing.Dict[str, str] mapping: The mapping of tokens used in the tree description to \
    valid taxon names.

    .. note::

        The ``TRANSLATE`` data is typically not accessed directly, but just used implicitly when
        calling :meth:`Trees.translate`.
    """
    def __init__(self, tokens, nexus=None):
        super().__init__(tokens, nexus=nexus)
        # get a word, and another word, then look for comma.
        self.mapping = collections.OrderedDict()
        key, value = None, None
        for t in iter_words_and_punctuation(self._tokens, nexus=nexus):
            if not key:
                assert isinstance(t, str)
                key = t
                continue
            if not value:
                assert isinstance(t, str)
                value = t
                continue
            assert t.is_punctuation and t.text == ',', t.text
            self.mapping[key] = value
            key, value = None, None
        if key and value:
            self.mapping[key] = value


class Tree(Payload):
    """
    This command describes a tree. Tree descriptions are standard object definition commands.
    They use the familiar parenthesis notation, with node names, branch lengths, and comments
    following the established Newick tree standard (see Felsenstein, 1993).

    The label of the node is a NEXUS token that is a taxon's defined name, a taxon's number, a
    taxon's label from the translation table, or a clade's defined name. The label is optional for
    internal nodes that are not observed taxa; it is not optional for terminal nodes. Internal
    nodes that have no label are represented implicitly by the parentheses containing the list of
    subclades. If the name of a TAXSET is used, it is interpreted as a list of the terminal taxa
    defined to be in the TAXSET (with commas implicitly inserted between the taxa). The length of
    the branch below the node is a number, positive or negative. Rooted and unrooted trees can be
    specified using the [&R] and [&U] comments at the start of the tree description. For example,

    .. code-block::

        TREE mytree = [&R] ((1,2),(3,4));

    is a rooted tree, whereas

    .. code-block::

        TREE mytree = [&U] ((1,2),(3,4));

    is an unrooted tree. The NEXUS standard does not specify whether rooted or unrooted is default.

    An example tree with branch lengths is

    .. code-block::

        TREE tree4 = ((beetle:4.3,fly:1.1):1.8,spider:2.5);

    If a file (and its data matrix) has four defined taxa, Crocodile, Bluebird, Archaeopteryx, and
    Rattlesnake, the following tree,

    .. code-block::

        TREE tree4= (((Bluebird)Archaeopteryx,Crocodile)Archosauria,Rattlesnake);

    would indicate that the taxon Archaeopteryx is ancestral to Bluebird and that Crocodile
    is their sister. Archosauria, because it does not refer to a taxon that has been defined
    in a TAXA or DATA block, is interpreted as the name of the clade including Archaeopteryx,
    Bluebird, and Crocodile. Any additional information about a clade, its ancestral node, or the
    branch below it is to be placed in NEXUS comment commands associated with the node. Al-
    though different programs may choose their own conventions for how to embed
    information in comments, the comments that begin with &N are reserved for future
    NEXUS comment commands. The NEXUS standard places no restrictions on the number of taxa
    contained in each tree.

    :ivar str name: The name of the tree.
    :ivar typing.Union[bool, None] rooted: Flag indicating whether the tree is rooted (or `None` \
    if no information is given)
    :ivar newick.Node newick: The tree description as `newick.Node`.

    .. code-block:: python

        >>> tree = Tree('tree4= (((Bluebird)Archaeopteryx,Crocodile)Archosauria,Rattlesnake)')
        >>> tree.name
        'tree4'
        >>> print(tree.newick.ascii_art())
                                        ┌─Archaeopteryx ──Bluebird
                        ┌─Archosauria───┤
        ────────────────┤               └─Crocodile
                        └─Rattlesnake
    """
    __multivalued__ = True

    def __init__(self, tokens, nexus=None):
        super().__init__(tokens, nexus=nexus)
        # We parse tree name and rooting information right away.
        self.name, ncomplete, e, self._rooted, nwk = None, False, False, None, False

        tokens = iter(self._tokens)
        while not self.name:
            t = next(tokens)
            if t.type in {TokenType.WORD, TokenType.QWORD}:
                self.name = t.text
                if t.type == TokenType.QWORD:
                    ncomplete = True

        while not e:
            t = next(tokens)
            if t.type == TokenType.WORD:
                assert not ncomplete, 'Stuff between tree name and ='
                self.name += t.text
            if t.type == TokenType.PUNCTUATION:
                if t.text == '=':
                    e = True
                else:
                    assert not ncomplete, 'Stuff between tree name and ='
                    self.name += t.text  # FIXME: Should we append punctuation to tree names?

        while not nwk:
            t = next(tokens)
            if t.type == TokenType.COMMENT and t.text.startswith('&'):
                self._rooted = t.text
            if t.type == TokenType.PUNCTUATION and t.text == '(':
                nwk = True
        assert nwk
        # Since Newick node construction is somewhat expensive, we defer it to lazy properties.
        self.newick_tokens = [Token(text='(', type=TokenType.PUNCTUATION)] + [t for t in tokens]
        self._nn = None

    @staticmethod
    def format(name: str,
               newick_node: newick.Node,
               rooted: typing.Optional[bool] = None) -> str:
        """
        Returns a representation of a tree as NEXUS string, suitable as payload of a ``TREE``
        command.
        """
        return '{} = {}{}'.format(
            Word(name).as_nexus_string(),
            '' if rooted is None else '[&{}] '.format('R' if rooted else 'U'),
            newick_node.newick)

    @property
    def rooted(self) -> typing.Union[None, bool]:
        """
        Whether the tree is rooted (`True`) or not (`False`) or no information is given (`None`).
        """
        if self._rooted:
            return self._rooted == '&R'

    @functools.cached_property
    def newick_string(self) -> str:
        """
        The Newick-formatted string representation of the tree.

        .. note::

            This property is intended for cases where only the string representation is of interest
            and the somewhat expensive construction of a `newick.Node` object is not necessary.
            Accessing the :meth:`Tree.newick` property will trigger node construction.

        .. warning::

            Due to some normalization (e.g. of whitespace) done by the Newick parser,
            `newick_string` may differ from `newick.newick`.

            .. code-block:: python

                >>> from commonnexus import Nexus
                >>> nex = Nexus('#nexus begin trees; tree 1 = (a,b)\\nc; end;')
                >>> nex.TREES.TREE.newick_string
                '(a,b)\\nc;'
                >>> nex.TREES.TREE.newick.newick
                '(a,b)c'
        """
        return ''.join(str(t) for t in self.newick_tokens) + ';'

    @functools.cached_property
    def newick(self) -> newick.Node:
        """
        A `newick.Node` instance parsed from the Newick representation of the tree.

        .. code-block:: python

            >>> from commonnexus import Nexus
            >>> nex = Nexus('#nexus begin trees; tree 1 = ((a,b)c,d)e; end;')
            >>> print(nex.TREES.TREE.newick.ascii_art())
                    ┌─a
                ┌─c─┤
            ──e─┤   └─b
                └─d
        """
        if self.nexus and self.nexus.cfg.validate_newick:
            # More correct, but slower: Let the newick parser validate the data.
            return newick.loads(self.newick_string)[0]
        # Quicker but by-passes some validation: Instantiate NewickString from pre-parsed Nexus
        # tokens!
        nt = [newick.Token('(', newick.TokenType.OBRACE, 0)]
        word, level = [], 1

        for token in self.newick_tokens[1:]:
            # now we assemble newick string and newick tokens in one go.
            if token.type == TokenType.WORD:
                word.append(token.text)
            elif token.type == TokenType.PUNCTUATION:
                if token.text in newick.RESERVED_PUNCTUATION:
                    # delimits words!
                    if word:
                        nt.append(newick.Token(''.join(word), newick.TokenType.WORD, level))
                        word = []
                    if token.text == ')':
                        level -= 1
                    nt.append(
                        newick.Token(token.text, newick.RESERVED_PUNCTUATION[token.text], level))
                    if token.text == '(':
                        level += 1
                else:
                    word.append(token.text)
            elif token.type == TokenType.COMMENT:
                if word:
                    nt.append(newick.Token(''.join(word), newick.TokenType.WORD, level))
                    word = []
                nt.append(newick.Token('[' + token.text + ']', newick.TokenType.COMMENT, level))
            elif token.type == TokenType.QWORD:
                assert not word  # As in NEXUS, a newick QWORD can not follow a WORD.
                nt.append(
                    newick.Token(Word(token.text).as_nexus_string(), newick.TokenType.QWORD, level))
        if word:
            nt.append(newick.Token(''.join(word), newick.TokenType.WORD, level))
        return newick.NewickString(nt).to_node()


class Trees(Block):
    """
    This block stores information about trees. The syntax for the TREES block is

    .. rst-class:: nexus

        | BEGIN TREES;
        |     [:class:`TRANSLATE <Translate>` arbitrary-token-used-in-tree-description
        |       valid-taxon-name [, arbitrary-token-used-in-tree-description valid-taxon-name...];]
        |     [:class:`TREE <Tree>` [*] tree-name= tree-specification;]
        | END;

    A :class:`TRANSLATE <Translate>` command, if present, must precede any
    :class:`TREE <Tree>` command.
    """
    __commands__ = [Translate, Tree]

    @property
    def trees(self) -> typing.List[Tree]:
        """
        Since TREE is one of the few NEXUS commands which may appear multiple times per block, we
        provide a shortcut to this list.
        """
        return self.commands['TREE']

    def validate(self, log=None):
        super().validate(log=log)
        valid, with_translate, tree_seen = True, False, False
        for i, cmd in enumerate(self[1:-1]):
            if cmd.name not in self.payload_map:  # pragma: no cover
                valid = log_or_raise(
                    'Invalid command for {} block: {}'.format(self.name, cmd.name), log=log)
            if cmd.name == 'TRANSLATE':
                if with_translate:  # pragma: no cover
                    valid = log_or_raise('Duplicate TRANSLATE command', log=log)
                elif tree_seen:  # pragma: no cover
                    valid = log_or_raise(
                        'TRANSLATE command **after** TREE command', log=log, level='warning')
                else:
                    with_translate = True
            else:
                tree_seen = True
        return valid

    @functools.cached_property
    def translate_mapping(self):
        mapping = {}
        if 'TRANSLATE' in self.commands:
            mapping.update(self.TRANSLATE.mapping)
        if 'TAXA' in self.linked_blocks:
            mapping.update({
                str(k): v for k, v in self.linked_blocks['TAXA'].TAXLABELS.labels.items()})
        elif self.nexus.TAXA and self.nexus.TAXA.TAXLABELS:
            mapping.update({str(k): v for k, v in self.nexus.TAXA.TAXLABELS.labels.items()})
        return mapping

    def translate(self, tree: typing.Union[Tree, newick.Node]) -> newick.Node:
        """
        Translate a tree according to the mapping TREES TRANSLATE.

        :return: A Newick node where the node labels have been translated to valid taxon labels.

        .. note::

            Translating a tree does **not** change tree's representation in the containing
            ``Nexus`` instance. To replace un-translated trees in a NEXUS file with translated
            ones, the following code would work:

            .. code-block:: python

                >>> untranslated = Nexus.from_file(path)
                >>> trees = []
                >>> for tree in untranslated.TREES.trees:
                ...     trees.append(Tree.format(
                ...         tree.name,
                ...         untranslated.TREES.translate(tree).newick,
                ...         rooted=tree.rooted))
                >>> untranslated.replace_block(
                ...     untranslated.TREES, [('TREE', tree) for tree in trees])
                >>> path.write_text(str(untranslated))
        """
        res = (tree.newick if isinstance(tree, Tree) else tree).rename(
            auto_quote=True, **self.translate_mapping)
        if not set(n.unquoted_name for n in res.walk() if n.name and n.is_leaf).issubset(
                self.translate_mapping.values()):
            warnings.warn('un-translatable leaf nodes!')
        return res

    @classmethod
    def from_data(cls,
                  *tree_specs: typing.Iterable[TreeSpec],
                  **translate_labels: typing.Dict[str, str]) -> 'Trees':
        """
        Create a TREES block from a list of tree specifications.

        If `translate_labels` are passed in, a corresponding TRANSLATE command will be added to the
        block and the trees will be "de-translated" accordingly.
        """
        TITLE = translate_labels.pop('TITLE', None)
        LINK = translate_labels.pop('LINK', None)
        ID = translate_labels.pop('ID', None)
        nexus = translate_labels.pop('nexus', None)
        cmds = []

        if translate_labels:
            cmds.append((
                'TRANSLATE',
                ',\n'.join('{} {}'.format(
                    Word(k).as_nexus_string(),
                    Word(v).as_nexus_string())
                    for k, v in sorted(translate_labels.items()))
            ))
        for name, nwk, rooted in tree_specs:
            if isinstance(nwk, str):
                nwk = newick.loads(nwk)[0]
            if translate_labels:
                nwk.rename(auto_quote=True, **{v: k for k, v in translate_labels.items()})
            cmds.append(('TREE', Tree.format(name, nwk, rooted)))
        return cls.from_commands(cmds, nexus=nexus, TITLE=TITLE, LINK=LINK, ID=ID)



from .base import Block, Payload


class Charset(Payload):
    """
    This command specifies and names a set of characters; this name can then be used in subsequent
    CHARSET definitions or wherever a character-set is required. The VECTOR format consists of 0's
    and 1's: a 1 indicates that the character is to be included in the CHARSET; whitespace is not
    necessary between 0's and l's. The name of a CHARSET cannot be equivalent to a character name
    or character number.

    **Predefined character-sets**:

    - The character-set ``CONSTANT`` is predefined for all DATATYPES; it specifies all invariant
      characters.
    - The character-set ``REMAINDER`` is predefined for all DATATYPES; it specifies all characters
      not previously referenced in the command.
    - The character-set ``GAPPED`` is predefined for all DATATYPES; it specifies all characters
      with a gap for at least one taxon.

    There are four additional predefined character-sets for characters of DATATYPE=DNA, RNA, and
    NUCLEOTIDE:

    1. ``POS1`` - All characters defined by current CODONPOSSET as first positions.
    2. ``POS2`` - All characters defined by current CODONPOSSET as second positions.
    3. ``POS3`` - All characters defined by current CODONPOSSET as third positions.
    4. ``NONCODING`` - All characters defined by current CODONPOSSET as non-protein-coding sites.
    """
    # syntax: a b c-g x-.


class Stateset(Payload):
    """
    This command allows one to name a set of states; it is not currently supported by any program.
    It is not available for DATATYPE=CONTINUOUS. For STANDARD format, the state-set is described by
    a list of state symbols, except that it should not be enclosed in parentheses or braces. Any
    current state-set symbols are valid in the state-set description.
    The following STATESET

    .. code-block::

        STATESET theSet = 2 3 4 5;

    defines the set composed of states 2, 3, 4, and 5.

    The VECTOR format consists of 0's and 1's: a 1 indicates that the state is to be included in
    the STATESET; whitespace is not necessary between 0's and l's. For example, the state-set

    .. code-block::

        STATESET theSet (VECTOR) =1001000;

    designates theSet to be the set containing first and fourth states.

    .. warning::

        `commonnexus` can read NEXUS containing this command, but will not resolve references to
        state-sets anywhere.
    """


class Changeset(Payload):
    """
    This command allows naming of a set of state changes; it is not currently supported by any
    program. It is not available for DATATYPE=CONTINUOUS. The description of the CHANGESET consists
    of pairs of state-sets joined by an operator. State-sets that consist of more than one token
    must be contained in parentheses. There are two allowed operators: ``->`` and ``<->``
    (``<-`` is **not** allowed). These operators can best be explained by example.

    .. code-block::

        CHANGESET changes1 = (1 2 3) -> (4 6) ;
        CHANGESET changes2 = 1 <-> 4 ;
        CHANGESET transversions = (A G) <-> (C T) ;

    The first CHANGESET represents any change from 1 to 4, 1 to 6, 2 to 4, 2 to 6, 3 to 4, or 3 to
    6, and the second set represents changes from 1 to 4 and 4 to 1. The CHANGESET "transversions"
    defines the set of all changes between purines and pyrimidines as transversions.
    """


class Taxset(Payload):
    """
    This command defines a set of taxa. A TAXSET name can be used in subsequent TAXSET definitions
    or wherever a taxon-set is required. The name of a TAXSET cannot be equivalent to a taxon name
    or taxon number. The taxa to be included are described in a taxon-set. For example, the
    following command

    .. code-block::

        TAXSET beetles=0mma-.;

    defines the TAXSET "beetles" to include all taxa from the taxon Omma to the last defined taxon.
    The VECTOR format consists of 0's and 1's: a 1 indicates that the taxon is to be included in
    the TAXSET; whitespace is not necessary between 0's and 1's.
    """


class Treeset(Payload):
    """
    This command defines a set of trees. A TREESET name can be used in subsequent TREESET
    definitions or wherever a tree-set is required. It is not currently supported by any program.
    It follows the same general format as a TAXSET command.

    .. warning::

        `commonnexus` can read NEXUS containing this command, but will not resolve references to
        tree-sets anywhere.
    """


class Partition(Payload):
    """
    [*]PARTITION commands define partitions of characters, taxa, and trees, respectively.
    The partition divides the objects into several (mutually exclusive) subsets.
    They all follow the same format. There are several formatting options. The VECTOR format
    consists of a list of partition names. By default, the name of each subset is a NEXUS word
    (this is the TOKENS option). The NOTOKENS option is only available in the VECTOR format; this
    allows use of single symbols for the subset names. Each value in a definition in VECTOR format
    must be separated by whitespace if the names are tokens but not if they are NOTOKENS. The
    following two examples are equivalent:

    .. code-block::

        TAXPARTITION populations = 1:1-3 , 2: 4-6 , 3:7 8;
        TAXPARTITION populations (VECTOR NOTOKENS) =11122233;

    The following two examples are equivalent:

    .. code-block::

        TAXPARTITION mypartition= Chiricahua: 1-3, Huachuca: 4-6, Galiuro: 7 8;
        TAXPARTITION mypartition (VECTOR) =
            Chiricahua Chiricahua Chiricahua Huachuca Huachuca Huachuca Galiuro Galiuro;
    """


class Charpartition(Partition):
    pass


class Taxpartition(Partition):
    pass


class Treepartition(Partition):
    pass


class Sets(Block):
    """
    This block stores sets of objects (characters, states, taxa, etc.).

    The general structure of the SETS block is as follows.

    .. rst-class:: nexus

        | BEGIN SETS;
        |     [:class:`CHARSET <Charset>` charset-name [({STANDARD | VECTOR})] = character-set; ]
        |     [:class:`STATESET <Stateset>` stateset-name [({STANDARD | VECTOR})] = state-set;]
        |     [:class:`CHANGESET <Changeset>`
        |         changeset-name=state-set<-> state-set [state-set<-> state-set...];]
        |     [:class:`TAXSET <Taxset>` taxset-name [({ STANDARD | VECTOR})] = taxon-set; ]
        |     [:class:`TREESET <Treeset>` treeset-name [({STANDARD | VECTOR})] = tree-set;]
        |     [:class:`CHARPARTITION <Charpartition>` partition-name
        |         [([{[NO]TOKENS}] [{STANDARD | VECTOR}])]
        |         =subset-name:character-set[, subset-name:character-set ...]
        |     ;]
        |     [:class:`TAXPARTITION <Taxpartition>` partition-name
        |         [([{[No]TOKENS}] [{STANDARD | VECTOR}])]
        |         =subset-name:taxon-set[, subset-name:taxon-set...]
        |     ;]
        |     [:class:`TREEPARTITION <Treepartition>` partition-name
        |         [([{[No]TOKENS}] [{STANDARD | VECTOR}])]
        |         =subset-name: tree-set[, subset-name:tree-set...]
        |     ;]
        | END;

    An example SETS block is

    .. code-block::

        BEGIN SETS;
            CHARSET larval = 1-3 5-8;
            STATESET eyeless = 0;
            STATESET eyed = 1 2 3;
            CHANGESET eyeloss = eyed -> eyeless;
            TAXSET outgroup=l-4;
            TREESET AfrNZVicariance = 3 5 9-12;
            CHARPARTITION bodyparts=head: 1-4 7, body:5 6, legs: 8-10 ;
        END;
    """
    __commands__ = [
        Charset, Stateset, Changeset, Taxset, Treeset, Charpartition, Taxpartition, Treepartition]



import itertools

from commonnexus.tools import combine
from commonnexus.cli_util import add_nexus


def help():
    return combine.__doc__


def register(parser):
    add_nexus(parser, many=True)
    parser.add_argument(
        '--drop-unsupported',
        help='Drop NEXUS blocks that cannot be combined.',
        action='store_true',
        default=False)


def run(args):
    kw = {}
    if args.drop_unsupported:
        kw['drop_unsupported'] = True
    print(combine.combine(*list(itertools.chain(*args.nexus)), **kw))



"""
Utility functions used to implement `commonnexus` subcommands.
"""
import re
import sys
import typing
import pathlib
import argparse

from commonnexus import Nexus

NEXUS = re.compile(r'(?:^|(?<=;|\s))(#nexus)', re.IGNORECASE | re.MULTILINE)


class ParserError(Exception):
    pass


class NexusType:
    """
    :class:`Ǹexus` as `Argument type <https://docs.python.org/3/library/argparse.html#type>`_

    Can create :class:`Ǹexus` instances from

    - file paths
    - string content
    - data read from `stdin`
    """
    def __init__(self, many: bool = False):
        self._many = many

    def __call__(self, string):
        if pathlib.Path(string).exists():
            nex = Nexus.from_file(string)
            return [nex] if self._many else nex
        if string == '-':
            string = sys.stdin.read()

        if self._many:
            res = []
            chunks = NEXUS.split(string)
            for i in range(1, len(chunks), 2):
                res.append(Nexus(chunks[i] + chunks[i + 1]))
            return res
        return Nexus(string)


def add_nexus(parser, many=False):
    """
    Adds (a) positional argument(s) `nexus` to the argument parser.

    :param parser:
    :param many:

    .. code-block:: python

        >>> import argparse
        >>> from commonnexus.cli_util import add_nexus
        >>> parser = argparse.ArgumentParser()
        >>> add_nexus(parser)
        >>> args = parser.parse_args(['#nexus begin block; end;'])
        >>> print(args.nexus.BLOCK)
        begin block; end;
        >>> parser = argparse.ArgumentParser()
        >>> add_nexus(parser, many=True)
        >>> args = parser.parse_args(['#nexus begin block; end;', '#nexus begin other; end;'])
        >>> len(args.nexus)
        2
    """
    if many:
        parser.add_argument(
            'nexus',
            help='NEXUS content specified as file path or string or "-" to read from stdin. Note '
                 'that "-" can only be used once. Content from a string or stdin will be split '
                 'into individual NEXUS "files" using the "#NEXUS" token as separator.',
            nargs='+',
            type=NexusType(many=True))
    else:
        parser.add_argument(
            'nexus',
            type=NexusType())


def add_flag(parser, name, help):
    parser.add_argument('--{}'.format(name), action='store_true', default=False, help=help)


def add_rename(parser, what):
    parser.add_argument(
        '--rename',
        help="Rename a {0} specified as 'old,new' where 'old' is the current name or number and "
             "'new' is the new name or as Python lambda function accepting a {0} label as input."
             "".format(what),
        type=lambda s: eval(s) if s.startswith('lambda') else tuple(s.split(',', maxsplit=1)),
        default=None)


def list_of_ranges(dstring) -> typing.List[int]:
    """
    Converts a comma-separated list of 1-based ranges into a list of 1-based indices.

    :param dstring: A string

    :return: `list` of zero-based indices
    """
    def _int(v):
        try:
            return int(v)
        except ValueError:  # pragma: no cover
            raise argparse.ArgumentTypeError("%r is not an integer" % v)

    out = []
    for token in dstring.split(','):
        token = token.replace(':', '-')
        if '-' in token:
            start, stop = token.split("-")
            out.extend([x for x in range(_int(start), _int(stop) + 1)])
        else:
            out.append(_int(token))
    return sorted(out)


def validate_operations(ops, args):
    if sum(1 for op in ops if getattr(args, op.name)) > 1:
        raise ParserError("Only one operation can be specified")  # pragma: no cover



import types
import typing
import warnings
import functools
import collections

from .base import Block, Payload
from commonnexus.util import log_or_raise
from commonnexus.tokenizer import (
    iter_words_and_punctuation, Token, iter_delimited, iter_lines, BOOLEAN, word_after_equals, Word,
)
from .taxa import Taxlabels

# A state in a CHARACTERS matrix may be missing, gapped, a state symbol or uncertain/polymorphic
# states.
State = typing.Union[None, str, typing.Set[str], typing.Tuple[str]]
StateMatrix = typing.OrderedDict[str, typing.OrderedDict[str, State]]

GAP = '\uFFFD'  # REPLACEMENT CHARACTER used to replace an [...] unrepresentable character
#: Some - but not all - punctuation is invalid as (special) state symbol.
INVALID_SYMBOLS = "()[]{}/\\,;:=*'\"*`<>^"


def duplicate_charlabel(label, cmd, nexus):
    if nexus and nexus.cfg.strict:  # pragma: no cover
        raise ValueError('character names must be unique!')
    else:
        warnings.warn(
            'Duplicate character name "{}" in {} command'.format(label, cmd))


class Eliminate(Payload):
    """
    This command allows specification of a list of characters that are to be excluded from
    consideration. Programs are expected to ignore ELIMINATEd characters completely during reading.
    In avoiding allocation of memory to store character information, the programs can save a
    considerable amount of computer memory. (This subcommand is similar to ZAP in version 3.1.1
    of PAUP.) For example,

    .. code-block::

        ELIMINATE 4-100;

    tells the program to skip over characters 4 through 100 in reading the matrix. Character-set
    names are not allowed in the character list. This command does not affect character numbers.

    .. warning:: The ``ELIMINATE`` command is currently not supported in `commonnexus`.
    """
    def __init__(self, tokens, nexus=None):
        super().__init__(tokens, nexus=nexus)
        if nexus is not None and not nexus.cfg.ignore_unsupported:
            raise NotImplementedError('The ELIMINATE command is not supported')


class Dimensions(Payload):
    """
    The DIMENSIONS command specifies the number of characters. The number following NCHAR
    is the number of characters in the data matrix. The NEXUS standard does not impose limits on the
    number of characters; a limit may be imposed by particular computer programs.

    It is strongly advised that new taxa not be defined in a CHARACTERS block, for the reasons
    discussed in the description of the DATA block. If new taxa are to be defined, this must be
    indicated by the NEWTAXA subcommand, specifying that new taxa are to be defined (this allows
    the computer program to prepare for creation of new taxa). NEWTAXA, if present, must appear
    before the NTAX subcommand. The NTAX subcommand, indicating the number of taxa in the MATRIX
    command in the block, is optional, unless NEWTAXA is specified, in which case it is required.

    :ivar bool newtaxa:
    :ivar typing.Optional[int] ntax:
    :ivar int nchar:
    """
    def __init__(self, tokens, nexus=None):
        super().__init__(tokens, nexus=nexus)
        self.newtaxa = False
        self.ntax = None
        self.char = None
        words = iter_words_and_punctuation(self._tokens, nexus=nexus)
        while 1:
            try:
                word = next(words)
                subcommand = None
                if isinstance(word, str):
                    subcommand = word.upper()
                if subcommand == 'NEWTAXA':
                    self.newtaxa = True
                elif subcommand == 'NTAX':
                    n = next(words)
                    assert n.text == '='
                    self.ntax = int(next(words))
                elif subcommand == 'NCHAR':
                    n = next(words)
                    assert n.text == '='
                    self.nchar = int(next(words))
            except StopIteration:
                break
        self.check()

    def check(self):
        assert self.nchar and ((not self.newtaxa) or self.ntax)


class Format(Payload):
    """
    The FORMAT command specifies the format of the data MATRIX. This is a crucial command because
    misinterpretation of the format of the data matrix could lead to anything from incorrect results
    to spectacular crashes. The DATATYPE subcommand must appear first in the command.

    The RESPECTCASE subcommand must appear before the MISSING, GAP, SYMBOLS, and MATCHCHAR
    subcommands.

    The following are possible formatting subcommands.

    1. DATATYPE = {STANDARD | DNA | RNA | NUCLEOTIDE | PROTEIN | CONTINUOUS}.
       This subcommand specifies the class of data. If present, it must be the first subcommand
       in the FORMAT command. Standard data consist of any general sort of discrete character data,
       and this class is typically used for morphological data, restriction site data, and so on.
       DNA, RNA, NUCLEOTIDE, and PROTEIN designate molecular sequence data. Meristic morphometric
       data and other information with continuous values can be housed in matrices of
       DATATYPE=CONTINUOUS. These DATATYPES are described in detail, with examples, at the end of
       the description of the CHARACTERS block.

       .. warning::

        ``DATATYPE=CONTINUOUS`` is currently not supported in `commonnexus`.
        Some programs accept (or expect) datatypes beyond the ones defined in the NEXUS spec;
        e.g. MrBayes has ``DATATYPE=RESTRICTION`` and Beauti may create "NEXUS" files with
        ``DATATYPE=BINARY``. `commonnexus` does not accept these non-standard datatypes and raises
        an exception when trying to read the MATRIX. Thus, to make "NEXUS" files with
        non-standard datatypes readable for `commonnexus`, substituting ``DATATYPE=STANDARD`` is
        typically the right thing to do.

    2. RESPECTCASE. By default, information in a MATRIX may be entered in uppercase, lowercase, or
       a mixture of uppercase and lowercase. If RESPECTCASE is requested, case is considered
       significant in SYMBOLS, MISSING, GAP, and MATCHCHAR subcommands and in subsequent references
       to states. For example, if RESPECTCASE is invoked, then SYMBOLS="A a B b" designates four
       states whose symbols are A, a, B, and b, which can then each be used in the MATRIX command
       and elsewhere. If RESPECTCASE is not invoked, then A and a are considered homonymous state
       symbols. This subcommand must appear before the SYMBOLS subcommand. This subcommand is not
       applicable to DATATYPE = DNA, RNA, NUCLEOTIDE, PROTEIN, and CONTINUOUS.
    3. MISSING. This subcommand declares the symbol that designates missing data.
       The default is "?". For example, MISSING=X defines an X to represent missing
       data. Whitespace is illegal as a missing data symbol, as are the :data:`INVALID_SYMBOLS`
    4. GAP. This subcommand declares the symbol that designates a data gap (e.g., base absent in
       DNA sequence because of deletion or an inapplicable character in morphological data). There
       is no default gap symbol; a gap symbol must be defined by the GAP subcommand before any gaps
       can be entered into the matrix. For example, GAP=- defines a hyphen to represent a gap.
       Whitespace is illegal as a gap symbol, as are the :data:`INVALID_SYMBOLS`
    5. SYMBOLS. This subcommand specifies the symbols and their order for character states used in
       the file (including in the MATRIX command). For example, SYMBOLS="0 1 2 3 4 5 6 7" designates
       numbers 0 through 7 as acceptable symbols in a matrix. The SYMBOLS subcommand is not allowed
       for DATATYPE=CONTINUOUS. The default symbols list differs from one DATATYPE to another, as
       described under state symbol in the Appendix. Whitespace is not needed between elements:
       SYMBOLS="012" is equivalent to SYMBOLS="0 1 2". For STANDARD DATATYPES, a SYMBOLS subcommand
       will replace the default symbols list of "0 1". For DNA, RNA, NUCLEOTIDE, and PROTEIN
       DATATYPES, a SYMBOLS subcommand will not replace the default symbols list but will add
       character-state symbols to the SYMBOLS list. The NEXUS standard does not define the position
       of these additional symbols within the SYMBOLS list. (These additional symbols will be
       inserted at the beginning of the SYMBOLS list in PAUP and at the end in MacClade. MacClade
       will accept additional symbols for PROTEIN but not DNA, RNA, and NUCLEOTIDE matrices.)

       .. warning::

            While the specification requires the content of the SYMBOLS subcommand to be enclosed in
            doublequotes, `commonnexus` also allows unquoted content; i.e. `SYMBOLS=01` is treated
            as equivalent to `SYMBOLS="01"`.

    6. EQUATE. This subcommand allows one to define symbols to represent one matrix entry. For
       example, EQUATE="E=(012)" means that each occurrence of E in the MATRIX command will be
       interpreted as meaning states 0, 1, and 2. The equate symbols cannot be any of the
       :data:`INVALID_SYMBOLS` or any or the currently defined MISSING, GAP, MATCHCHAR, or state
       SYMBOLS. Case is significant in equate symbols. That is, MISSING=? EQUATE="E=(012)e=?" means
       that E will be interpreted as 0, 1, and 2 and e will be interpreted as missing data.
    7. MATCHCHAR. This subcommand defines a matching character symbol. If this subcommand is
       included, then a matching character symbol in the MATRIX indicates that the states are
       equivalent to the states possessed by the first taxon listed in the matrix for that
       character. In the following matrix, the sequence for taxon 2 is GACTTTC:

       .. code-block::

            BEGIN DATA;
                DIMENSIONS NCHAR = 7;
                FORMAT DATATYPE=DNA MATCHCHAR=.;
                MATRIX
                    taxon_l GACCTTA
                    taxon_2 ...T..C
                    taxon_3 ..T.C..;
            END;

       Whitespace is illegal as a matching character symbol, as are the :data:`INVALID_SYMBOLS`

       .. warning::

            `commonnexus` uses `"."` as default MATCHCHAR. So if `"."` is used as a regular state
            symbol, the NEXUS must be read using the `no_default_matchchar` config option.

    8. [NO]LABELS. This subcommand declares whether taxon or character labels are to appear on the
       left side of the matrix. By default, they should appear. If NOLABELS is used, then no labels
       appear, but then all currently defined taxa must be included in the MATRIX in the order in
       which they were originally defined.
    9. TRANSPOSE. This subcommand indicates that the MATRIX is in transposed format, with each row
       of the matrix representing the information from one character and each column representing
       the information from one taxon. The following is an example of a TRANSPOSEd MATRIX:

       .. code-block::

            MATRIX
                character_1 101101
                character_2 011100
                character_3 011110;

    10. INTERLEAVE. This subcommand indicates that the MATRIX is in interleaved format, i.e., it is
        broken up into sections. If the data are not transposed, then each section contains the
        information for some of the characters for all taxa. For example, the first section might
        contain data for characters 1-50 for all taxa, the second section contains data for
        characters 51-100, etc. Taxa in each section must occur in the same order. This format is
        especially useful for molecular sequence data, where the number of characters can be large.
        A small interleaved matrix follows:

        .. code-block::

            MATRIX
                taxon_1 ACCTCGGC
                taxon_2 ACCTCGGC
                taxon_3 ACGTCGCT
                taxon_4 ACGTCGCT
                taxon_1 TTAACGA
                taxon_2 TTAACCA
                taxon_3 CTCACCA
                taxon_4 TTCACCA

        The interleaved sections need not all be of the same length. In an interleaved matrix,
        newline characters are significant: they indicate that the next character information
        encountered applies to a different taxon (for nontransposed matrices).
    11. ITEMS. Each entry in the matrix gives information about a character's condition
        in a taxon. The ITEMS subcommand indicates what items of information are listed
        at each entry of the matrix. With discrete character data, the entry typically consists
        of the states observed in the taxon (either the single state observed or several states
        if the taxon is polymorphic or of uncertain state). This can be specified by the state-
        ment ITEMS=STATES, but because it is the default and the only option allowed by
        most current programs for discrete data, an ITEMS statement is usually unnecessary.
        For continuous data, however, the wealth of alternatives (average, median, variance,
        minimum, maximum, sample size)t often requires an explicit ITEMS statement to in-
        dicate what information is represented in each data matrix entry. Some ITEMS (such
        as VARIANCE) would be appropriate to only some DATATYPES; other ITEMS such as
        SAMPLESIZE and STATES would be appropriate to most or all DATATYPES. If more
        than one item is indicated, parentheses must be used to surround the list of items,
        e.g., ITEMS=(AVERAGE VARIANCE); otherwise the parentheses are unnecessary,
        e.g., ITEMS=AVERAGE. More information about ITEMS options can be found in the
        discussion of the different DATATYPES under MATRIX; information specifically about
        the STATES option is given under STATESFORMAT.

       .. warning::

            Settings other than ``ITEMS=STATES`` are currently not supported in `commonnexus`.

    12. STATESFORMAT. The entry in a matrix usually lists (for discrete data) or may list
        (for continuous data) the states observed in the taxon. The STATESFORMAT subcommand
        specifies what information is conveyed in that list of STATES. In most current programs for
        discrete data, when a taxon is polymorphic the entry of the matrix lists only what distinct
        states were observed, without any indication of the number or frequency of individuals
        sampled with each of the states. Thus, if all individuals sampled within the taxon have
        state A, the matrix entry would be A, whereas if some have state A and others have state B,
        the entry would be (AB), which corresponds to the option STATESFORMAT=STATESPRESENT.
        Because it is the default for discrete data, this statement is typically unnecessary with
        current programs. The other STATESFORMAT options can be illustrated with an example, in
        which two individuals of a taxon were observed to have state A and three were observed to
        have state B. When STATESFORMAT=INDIVIDUALS, the state of each of the individuals (or other
        appropriate sampling subunit) is listed exhaustively, (A A B B B); when STATESFORMAT=COUNT,
        the number of individuals with each observed state is indicated, e.g., (A:2 B:3); when
        STATESFORMAT=FREQUENCY, the frequencies of various observed states are indicated, e.g.,
        (A:0.40 B:0.60). The STATESFORMAT command may also be used for continuous data, for which
        the default is STATESFORMAT=INDIVIDUALS.

        .. warning::

            Only the default setting ``STATESFORMAT=STATESPRESENT`` is currently supported in
            `commonnexus`.

    13. [NO]TOKENS. This subcommand specifies whether data matrix entries are single symbols or
        whether they can be tokens. If TOKENS, then the data values must be full NEXUS tokens,
        separated by whitespace or punctuation as appropriate, as in the following example:

        .. code-block::

            BEGIN CHARACTERS;
                DIMENSIONS NCHAR= 3 ;
                CHARSTATELABELS 1 hair/absent
                    present, 2 color/red blue,
                    3 size/small big;
                FORMAT TOKENS;
                MATRIX
                    taxon_1 absent red big
                    taxon_2 absent blue small
                    taxon_3 present blue small ;
            END;

        TOKENS is the default (and the only allowed option) for DATATYPE=CONTINUOUS; NOTOKENS is
        the default for all other DATATYPES. TOKENS is not allowed for DATATYPES DNA, RNA, and
        NUCLEOTIDE. If TOKENS is invoked, the standard three-letter amino acid abbreviations can be
        used with DATATYPE=PROTEIN and defined state names can be used for DATATYPE=STANDARD.

        .. warning:: ``TOKENS`` is currently not supported in `commonnexus`.

    :ivar str datatype:
    :ivar bool respectcase:
    :ivar str missing:
    :ivar typing.Optional[str] gap:
    :ivar typing.List[str] symbols:
    :ivar typing.Dict[str, str] equate:
    :ivar typing.Optional[str] matchchar:
    :ivar typing.Optional[bool] labels:
    :ivar bool transpose:
    :ivar bool interleave:
    :ivar typing.List[str] items:
    :ivar typing.Optional[str] statesformat:
    :ivar typing.Optional[bool] tokens:

    .. note::

        It's typically not necessary to access the attributes of a ``Format`` instance from user
        code. Instead, the information is accessed when reading the matrix data in
        :meth:`Characters.get_matrix`.
    """
    def __init__(self, tokens, nexus=None):
        super().__init__(tokens, nexus=nexus)
        self.datatype = None
        self.respectcase = False
        self.missing = '?'
        self.gap = None
        self.symbols = ['0', '1']  # The default for DATATYPE=STANDARD
        self.equate = {}
        self.matchchar = None if nexus and nexus.cfg.no_default_matchchar else '.'
        self.labels = None
        self.transpose = False
        self.interleave = False
        self.items = []
        self.statesformat = None
        self.tokens = None
        self.explicit_symbols = False

        if tokens is None:
            return

        words = iter_words_and_punctuation(self._tokens, nexus=nexus)
        after_equals = functools.partial(word_after_equals, words)

        subcommand = None
        subcommands_set = set()
        while 1:
            try:
                word = next(words)
                if isinstance(word, str):
                    subcommand = word.upper()
                elif isinstance(word, Token) and word.text == '=':
                    if subcommand in ['RESPECTCASE', 'TRANSPOSE', 'INTERLEAVE', 'LABELS', 'TOKENS']:
                        # Some NEXUS variants set boolean subcommands always with "=no|yes"
                        word = next(words).lower()
                        if subcommand == 'LABELS' and word == 'left':
                            word = 'yes'
                        setattr(self, subcommand.lower(), BOOLEAN[word])
                        subcommands_set.add(subcommand)
                    elif subcommand:  # pragma: no cover
                        raise ValueError(subcommand)

                if subcommand in ['DATATYPE', 'MISSING', 'MATCHCHAR', 'GAP', 'STATESFORMAT']:
                    setattr(self, subcommand.lower(), after_equals())
                    if subcommand == 'DATATYPE' and self.datatype.upper() != 'STANDARD':
                        self.symbols = []
                elif subcommand in ['RESPECTCASE', 'TRANSPOSE', 'INTERLEAVE']:
                    if subcommand not in subcommands_set:
                        setattr(self, subcommand.lower(), True)
                elif subcommand in ['NOLABELS', 'LABELS', 'NOTOKENS', 'TOKENS']:
                    setattr(self, subcommand.replace('NO', '').lower(), 'NO' not in subcommand)
                elif subcommand == 'SYMBOLS':
                    self.explicit_symbols = True
                    self.symbols = []
                    next_token_text = after_equals()
                    if not next_token_text.startswith('"'):
                        self.symbols = list(next_token_text)
                    else:
                        for w in iter_delimited(next_token_text, words):
                            if isinstance(w, str):
                                self.symbols.extend(list(w))
                            else:
                                assert w.text in '+-'
                                self.symbols.append(w.text)
                elif subcommand == 'EQUATE':
                    key, e, bracket = None, False, None
                    for t in iter_delimited(after_equals(), words):
                        if isinstance(t, Token):
                            if t.text == '=':
                                assert key
                                e = True
                                bracket = None
                            else:
                                bracket = t.text
                        elif isinstance(t, str):
                            if key:
                                assert e
                                if bracket is None:
                                    assert len(t) == 1
                                    self.equate[key] = t
                                elif bracket == '(':
                                    self.equate[key] = tuple(t)
                                elif bracket == '{':
                                    self.equate[key] = set(t)
                                else:  # pragma: no cover
                                    raise ValueError(
                                        'Invalid punctuation in EQUATE content: {}'.format(bracket))
                                key, e = None, False
                            else:
                                key = t
                elif subcommand == 'ITEMS':
                    for w in iter_delimited(
                            after_equals(), words, delimiter='()', allow_single_word=True):
                        assert isinstance(w, str)
                        self.items.append(w)
            except StopIteration:
                break
        if self.datatype:
            self.datatype = self.datatype.upper()
            assert self.datatype in {
                'STANDARD', 'DNA', 'RNA', 'NUCLEOTIDE', 'PROTEIN', 'CONTINUOUS'}
            if self.datatype == 'CONTINUOUS' and not self.nexus.cfg.ignore_unsupported:
                raise NotImplementedError('DATATYPE=CONTINUOUS is not supported!')
        self.items = [i.upper() for i in self.items]
        assert all(
            i in 'MIN MAX MEDIAN AVERAGE VARIANCE STDERROR SAMPLESIZE STATES'.split()
            for i in self.items)
        if not self.items:
            self.items = ['STATES']
        if self.items != ['STATES']:
            raise NotImplementedError('Only ITEMS=STATES is supported!')
        if self.statesformat:
            self.statesformat = self.statesformat.upper()
            assert self.statesformat in {'STATESPRESENT', 'INDIVIDUALS', 'COUNT', 'FREQUENCY'}
        else:
            self.statesformat = 'STATESPRESENT'
        if self.statesformat != 'STATESPRESENT':
            raise NotImplementedError(
                'STATESFORMATs other than STATESPRESENT are not supported')
        for attr in ['missing', 'gap', 'matchchar']:
            c = getattr(self, attr)
            if c:
                assert len(c) == 1 and c not in INVALID_SYMBOLS
        if self.tokens:
            raise NotImplementedError('TOKENS is not supported')

        if self.datatype in {'DNA', 'RNA', 'NUCLEOTIDE'}:
            T = 'U' if self.datatype == 'RNA' else 'T'
            self.symbols.extend(list('ACG' + T))
            self.equate.update(
                R=set('AG'),
                Y=set('C' + T),
                M=set('AC'),
                K=set('G' + T),
                S=set('CG'),
                W=set('A' + T),
                H=set('AC' + T),
                B=set('CG' + T),
                V=set('ACG'),
                D=set('AG' + T),
                N=set('ACG' + T),
                X=set('ACG' + T),
            )
            if self.datatype == 'NUCLEOTIDE':
                self.equate.update(U='T')
        elif self.datatype == 'PROTEIN':
            self.symbols.extend(list('ACDEFGHIKLMNPQRSTVWY*'))
            self.equate.update(B=set('DN'), Z=set('EQ'))

        if not self.respectcase:
            self.equate = {k.upper(): v for k, v in self.equate.items()}

        invalid_equate = \
            list(INVALID_SYMBOLS) + self.symbols + \
            [self.missing or '', self.gap or '', self.matchchar or '']
        assert not any(c in invalid_equate for c in self.equate)


class Charstatelabels(Payload):
    """
    This command allows specification of both the names of the characters and the names of the
    states. This command was developed as an alternative to the older commands CHARLABELS and
    STATELABELS. For example,

    .. code-block::

        CHARSTATELABELS
            1 eye_color/red blue green,
            3 head_shape/round square,
            5 pronotum_size/small medium large

    A forward slash (/) separates the character name and the state names, with a comma separating
    the information for different characters. If no state names are to be specified, the slash may
    be omitted; if no character names are to be specified, the slash must be included, but no token
    needs to be included between the character number and the slash. If state x is the last state
    to be named, then subsequent states need not be named, but states 1 through x must be. If no
    name is to be applied to a state, enter a single underscore for its name. Character and state
    names are single NEXUS words. Character names must not correspond to another character name or
    number; thus, 1 is not a valid name for the second character listed. State names cannot be
    applied if DATATYPE=CONTINUOUS.

    :ivar typing.List[types.SimpleNamespace] characters:

    .. code-block:: python

        >>> cmd = Charstatelabels('1 eye_color/red blue green, 3 head_shape/round square')
        >>> cmd.characters[0].name
        'eye_color'
        >>> cmd.characters[0].states
        ['red', 'blue', 'green']

    .. warning::

        In strict mode (see :class:`commonnexus.nexus.Config`) duplicate character names will raise
        a ``ValueError``, otherwise a ``UserWarning`` will be emitted. While a matrix with duplicate
        character names can still be read, it will typically **not** be as expected, because only
        the values for the last character for a given name will be present.
    """
    def __init__(self, tokens, nexus=None):
        super().__init__(tokens, nexus=nexus)
        self.characters = []
        names = set()
        words = iter_words_and_punctuation(self._tokens, nexus=nexus)
        num, name, states, in_states, comma = None, None, [], False, False

        while 1:
            try:
                w = next(words)
                if num is None:
                    num = int(w)
                    continue
                if isinstance(w, Token) and w.text == ',':
                    comma = True  # We want to be able to detect trailing commas!
                    if name and name in names:
                        duplicate_charlabel(name, 'CHARSTATELABELS', nexus)
                    names.add(name)
                    self.characters.append(
                        types.SimpleNamespace(number=num, name=name, states=states))
                    num, name, states, in_states = None, None, [], False
                    continue
                if in_states:
                    states.append(w)
                    continue
                if isinstance(w, Token) and w.text == '/':
                    in_states = True
                    continue
                name = w
            except StopIteration:
                break
        if num:
            self.characters.append(types.SimpleNamespace(number=num, name=name, states=states))
        elif comma:  # There was a comma, but no new label.
            warnings.warn('Trailing comma in CHARSTATELABELS command')


class Charlabels(Payload):
    """
    This command allows specification of names of characters:

    .. code-block::

        CHARLABELS
            flange microsculpture
            body_length
            hind_angles #_spines
            spine_size _ _ head_size
            pubescent_intervals head_color
            clypeal_margin;

    Character labels are listed consecutively. If character x is the last character to be named,
    then subsequent characters need not be named, but characters 1 through x need to be. If no name
    is to be applied to a character, a single underscore can be used for its name. Character names
    are single NEXUS words. They must not correspond to another character name or number; thus, 1
    is not a valid name for the second character listed. The command should be used only for
    nontransposed matrices (in transposed matrices, the character labels are defined in the MATRIX
    command). We recommend that programs abandon this command in place of the more flexible
    CHARSTATELABELS command when writing NEXUS files, although programs should continue to read
    CHARLABELS because many existing NEXUS files use CHARLABELS.

    :ivar typing.List[types.SimpleNamespace] characters:
    """
    def __init__(self, tokens, nexus=None):
        super().__init__(tokens, nexus=nexus)
        self.characters = []
        names = set()
        for i, w in enumerate(iter_words_and_punctuation(self._tokens, nexus=nexus)):
            assert isinstance(w, str)
            if w and w in names:
                duplicate_charlabel(w, 'CHARLABELS', nexus)
            names.add(w)
            self.characters.append(types.SimpleNamespace(number=i + 1, name=w, states=[]))


class Statelabels(Payload):
    """
    This command allows specification of the names of states:

    .. code-block::

        STATELABELS
            1 absent present,
            2 isodiametric transverse,
            3 '4.5-6.2mm' '6.3-7.0mm' '7.7-11.0mm' '>12.0mm',
            4 rounded subangulate angulate,
            10 0 '1-4' '6-9' '7-9' '8-9' 7 8 9,
            11 black rufous metallic flavous,
            12 straight concave,

    State labels need not be specified for all characters. A comma must separate state labels for
    each character. State labels are listed consecutively within a character. If state x is the
    last state to be named, then subsequent states need not be named, but states 1 through x must
    be. If no name is to be applied to a state, enter a single underscore for its name. State
    names are single NEXUS words. This command is not valid for DATATYPE=CONTINUOUS.
    We recommend that programs abandon this command in place of the more flexible
    CHARSTATELABELS command when writing NEXUS files, although programs should continue to read
    STATELABELS because many existing NEXUS files use STATELABELS.

    :ivar typing.List[types.SimpleNamespace] characters:
    """
    def __init__(self, tokens, nexus=None):
        super().__init__(tokens, nexus=nexus)
        self.characters = []

        words = iter_words_and_punctuation(self._tokens, nexus=nexus)
        num, states = None, []

        while 1:
            try:
                w = next(words)
                if num is None:
                    num = int(w)
                    continue
                if isinstance(w, Token) and w.text == ',':
                    self.characters.append(
                        types.SimpleNamespace(number=num, name=None, states=states))
                    num, states = None, []
                    continue
                assert isinstance(w, str)
                states.append(w)
            except StopIteration:
                break
        if num and states:
            self.characters.append(types.SimpleNamespace(number=num, name=None, states=states))


class Matrix(Payload):
    """
    In its standard format, the MATRIX command contains a sequence of taxon names and state
    information for that taxon. The MATRIX itself is of the form

    .. code-block::

        MATRIX
            taxon-name entry entry... entry
            taxon-name entry entry... entry
            taxon-name entry entry... entry;

    Each entry in the matrix is the information about a particular character for a particular taxon.
    For example, it might be the assignment of state 0 to taxon 1 for character 1. Thus, the entry
    would consist of one state symbol, 0. If the taxon were polymorphic, the entry would consist
    of multiple state symbols, e.g. (0 1), indicating the taxon has both states 0 and 1. More
    details about the nature of each entry of the matrix are given under ITEMS and under each
    DATATYPE. Each entry needs to be enclosed in parentheses or braces whenever more than one state
    symbol is given, e.g. (01) with standard data and the default NOTOKENS option, or if the
    information is conveyed by more than one NEXUS token, e.g., (0:100) or (2.3 4.5 6.7). Otherwise,
    the parentheses or braces are optional. No whitespace is needed between entries in the matrix
    unless the TOKENS subcommand of the FORMAT command is invoked or implied and parentheses or
    braces do not surround an entry.
    Taxa need not be in the same order as in the TAXA block, and the matrix need not contain all
    taxa. For interleaved matrices, all sections must have the same taxa in the same order.
    Examples of matrices of different DATATYPES are described below.

    1. For STANDARD data, each entry of the matrix consists of a single state-set. Under the
       defaults (ITEMS=STATES and STATESFORMAT=STATESPRESENT), each entry of the matrix consists of
       a single state-set; if there are multiple states, then the entry must be enclosed in
       parentheses (indicating polymorphism) or braces (indicating uncertainty in state). For
       example, in the following matrix,

       .. code-block::

            BEGIN CHARACTERS;
                DIMENSIONS NCHAR=9;
                FORMAT SYMBOLS="-+x";
                MATRIX
                    taxon_1 (-+){-+}+---+--
                    taxon_2 +x-++--+x
                    taxon_3 -++++--+x;
            END;

       taxon_1 is polymorphic for the first character and has either state - or state + for the
       second character. If STATESFORMAT=COUNT or FREQUENCY, then each entry must be enclosed in
       parentheses because more than one token is required to convey information for even one state:

       .. code-block::

            BEGIN CHARACTERS;
                DIMENSIONS NCHAR=3;
                FORMAT STATESFORMAT=FREQUENCY SYMBOLS = "012";
                MATRIX
                    taxon_1 (0:0.251:0.75) (0:0.31:0.7) (0:0.51:0.32:0.2)
                    taxon_2 (0:0.41:0.6) (0:0.81:0.2) (1:0.152:0.85)
                    taxon_3 (0:0.01:1.0) (0:0.551:0.45) (0:0.11:0.9);
            END;

    2. For DNA, RNA, NUCLEOTIDE, and PROTEIN data, each entry of the matrix consists of one or more
       state symbols describing the state(s) at one site in a molecular sequence. If
       STATESFORMAT=STATESPRESENT and if an entry represents a single state, then it is represented
       as a single state symbol (or if DATATYPE=PROTEIN and TOKENS, as a three-letter amino acid
       name). If an entry represents multiple states, then it must be enclosed in parentheses
       (indicating polymorphism) or braces (indicating uncertainty in state). Following is a matrix
       of DATATYPE=DNA:

       .. code-block::

            BEGIN CHARACTERS;
                DIMENSIONS NCHAR=12;
                FORMAT DATATYPE = DNA;
                MATRIX
                    taxon_1 ACCATGGTACGT
                    taxon_2 TCCATGCTACCC
                    taxon_3 TCCATGGAACCC;
            END;

    3. For CONTINUOUS data, each entry in the matrix must be enclosed by parentheses if more than
       one item is specified in the ITEMS subcommand. Parentheses must also be used whenever
       multiple tokens are needed for an entry in the matrix. If an entry consists of a single
       token (eg., 0.231), it may be written without parentheses but must then be separated from
       other entries by whitespace.

       .. code-block::

            MATRIX
                A 0.453 1.43 78.6
                B 0.34 1.02 55.7
                C 0.22 1.79 69.1;

       A matrix entry can include average, minimum, maximum, variance, standard error, sample size,
       and a listing of states observed in the taxon, as specified in the ITEMS subcommand. The
       sample size, if included, must be in the form of an integer; the other numbers can be either
       in English decimal (e.g., 0.00452) or in exponential form (e.g., 4.52E-3). The information
       listed for each taxon for a continuous character is specified in the ITEMS subcommand of the
       FORMAT command. For example, if the matrix contains only information about the minimum and
       maximum value for each taxon, the ITEMS subcommand would be ITEMS=(MIN MAX) and a small
       matrix might look something like this:

       .. code-block::

            MATRIX
                taxon_1 (0.21 0.45) (0.34 0.36)
                taxon_2 (0.13 0.22) (0.45 0.55);

       If the ITEMS include the raw measurements (states), e.g., to list a sample of measurements
       from individuals, then the other items must precede the listing of states. There is no
       restriction on the number of elements in the listing of states. This example has only one
       continuous character:

       .. code-block::

            FORMAT DATATYPE=CONTINUOUS ITEMS=(AVERAGE STATES) STATESFORMAT=INDIVIDUALS;
            MATRIX
                taxon_1 (1.2 2.1 1.6 0.8 1.8 0.3 0.6)
                taxon_2 (1.6 2.2 1.7 1.0 2.0 1.6 1.9 0.8);

       in which the first value is the sample average and the subsequent values comprise the sample
       of observed states. Possible ITEMS to be included are MIN (minimum), MAX (maximum), AVERAGE
       (sample average), VARIANCE (sample variance), STDERROR (standard error), MEDIAN
       (sample median), SAMPLESIZE, and STATES. The manner of presentations of states can be
       indicated using the STATESFORMAT command. The default ITEMS for continuous data is AVERAGE.

    .. note::

        Since reading the matrix data only makes sense if information from other commands - in
        particular :class:`FORMAT <Format>` - is considered, the ``Matrix`` object does not have
        any attributes for data access. Instead, the matrix data can be read via
        :meth:`Characters.get_matrix`.
    """


class Characters(Block):
    """
    A CHARACTERS block defines characters and includes character data.

    Taxa are usually not defined in a CHARACTERS block; if they are not, the CHARACTERS block must
    be preceded by a block that defines taxon labels and ordering (e.g., TAXA).

    Syntax of the CHARACTERS block is as follows:

    .. rst-class:: nexus

        | BEGIN CHARACTERS;
        |   :class:`DIMENSIONS <Dimensions>` [NEWTAXA NTAX=num-taxa] NCHAR=num-characters;
        |   [:class:`FORMAT <Format>`
        |       [DATATYPE = { STANDARD| DNA | RNA | NUCLEOTIDE | PROTEIN | CONTINUOUS} ]
        |       [RESPECTCASE]
        |       [MISSING=symbol]
        |       [GAP=symbol]
        |       [SYMBOLS="symbol [symbol...]"]
        |       [EQUATE="symbol = entry [symbol = entry... ] " ]
        |       [MATCHCHAR= symbol ]
        |       [[NO]LABELS]
        |       [TRANSPOSE]
        |       [INTERLEAVE]
        |       [ITEMS=([MIN] [MAX] [MEDIAN] [AVERAGE] [VARIANCE] [STDERROR] [SAMPLESIZE] [STATES])]
        |       [STATESFORMAT= {STATESPRESENT | INDIVIDUALS | COUNT | FREQUENCY}]
        |       [[NO]TOKENS]
        |   ;]
        |   [:class:`ELIMINATE <Eliminate>` character-set;]
        |   [:class:`TAXLABELS <commonnexus.blocks.taxa.Taxlabels>` taxon-name [taxon-name ...];]
        |   [:class:`CHARSTATELABELS <Charstatelabels>`
        |       character-number [character-name] [/state-name [state-name...]]
        |       [, character-number [character-name] [/state-name [state-name...]]...]
        |   ;]
        |   [:class:`CHARLABELS  <Charlabels>` character-name [character-name...];]
        |   [:class:`STATELABELS <Statelabels>`
        |       character-number [state-name [state-name ...]]
        |       [, character-number [state-name [state-name...]]...]
        |   ;]
        |   :class:`MATRIX <Matrix>` data-matrix;
        | END;

    :class:`DIMENSIONS <Dimensions>`, :class:`FORMAT <Format>`, and :class:`ELIMINATE <Eliminate>`
    must all precede :class:`CHARLABELS <Charlabels>`, :class:`CHARSTATELABELS <Charstatelabels>`,
    :class:`STATELABELS <Statelabels>`, and :class:`MATRIX <Matrix>`.
    :class:`DIMENSIONS <Dimensions>` must precede :class:`ELIMINATE <Eliminate>`.
    Only one of each command is allowed per block.
    """
    __commands__ = [
        Dimensions, Format, Eliminate, Taxlabels, Charstatelabels, Charlabels, Statelabels, Matrix]

    def is_binary(self) -> bool:
        """
        :return: Whether the matrix in the block is binary, i.e. codes items as presence/absence \
        using symbols "01".
        """
        return not bool(self.FORMAT) or (self.FORMAT.symbols == ['0', '1'])

    def get_matrix(self, labeled_states: bool = False) -> StateMatrix:
        """
        :param labeled_states: Flag signaling whether state symbols should be translated to state \
        labels (if available).
        :return: The values of the matrix, read according to FORMAT. The matrix is returned as \
        ordered `dict`, mapping taxon labels (if available, else numbers) to ordered `dict`s \
        mapping character labels (if available, else numbers) to state values. State values are \
        either atomic values (of type `str`) or `tuple`s (indicating polymorphism) or `set`s \
        (indicating uncertainty) of atomic values. Atomic values may be `None` (indicating missing \
        data), the special string `GAP` (indicating gaps) or state symbols or labels (if available \
        and explicitly requested via `labeled_states=True`). State symbols are returned using the \
        case given in FORMAT SYMBOLS, i.e. if a RESPECTCASE directive is missing and \
        FORMAT SYMBOLS="ABC", a value "a" in the matrix will be returned as "A".
        """
        format = Format(None) if 'FORMAT' not in self.commands else self.FORMAT

        # Determine dimensions and labels:
        ntax, taxlabels = self.get_taxlabels(format)
        nchar = self.DIMENSIONS.nchar
        charlabels, statelabels = self.get_charstatelabels(nchar, format)
        if format.transpose and (not format.interleave) and (format.labels != False) and (not ntax):
            raise ValueError("Can't read transposed matrix without NTAX.")  # pragma: no cover
        if format.datatype == 'CONTINUOUS':  # pragma: no cover
            raise NotImplementedError("Can't read a matrix of datatype CONTINUOUS")

        # We read the matrix data in an agnostic way, ignoring whether it's transposed or not, as
        # ordered dictionary mapping row labels (or numbers) to lists of entries.
        res = collections.OrderedDict()
        label, entries = None, []
        ncols, nrows = ntax if format.transpose else nchar, nchar if format.transpose else ntax

        for i, line in enumerate(
                list(iter_lines(self.MATRIX._tokens)) if format.interleave else
                [self.MATRIX._tokens],
                start=1):
            words = iter_words_and_punctuation(
                line, allow_punctuation_in_word='+-', nexus=self.nexus)
            while 1:
                try:
                    t = next(words)
                    if (format.labels is not False) and label is None:
                        assert isinstance(t, str)
                        label = t
                        continue
                    if isinstance(t, Token):
                        if t.text == '(':
                            w = next(words)
                            symbols = ''
                            while isinstance(w, str) or (w.text in format.symbols) \
                                    or (w.text == ","):
                                if isinstance(w, str) or w.text != ',':
                                    symbols += getattr(w, 'text', w)
                                w = next(words)
                            assert w.text == ')', "Expected )"
                            entries.append(tuple(symbols))
                        elif t.text == '{':
                            w = next(words)
                            vals = set()
                            while isinstance(w, str) or (w.text in format.symbols) \
                                    or (w.text == format.gap) or (w.text == ","):
                                if isinstance(w, str) or w.text != ',':
                                    vals |= set(getattr(w, 'text', w))
                                w = next(words)
                            assert w.text == '}', "Expected }"
                            entries.append(vals)
                        elif t.text in format.symbols:  # pragma: no cover
                            entries.append(t.text)
                        else:  # pragma: no cover
                            raise ValueError('Unexpected punctuation in matrix')
                    else:
                        entries.extend(list(t))  # We split a word into a list of symbols.
                    if not format.interleave and (len(entries) == ncols):
                        res[label or (len(res) + 1)] = entries
                        label, entries = None, []
                except StopIteration:
                    break
            if format.interleave:
                key = label or (i % nrows or nrows)
                if key not in res:
                    res[key] = []
                res[key].extend(entries)
                label, entries = None, []

        cols = ncols or len(list(res.values())[0])
        assert all(len(states) == cols for states in res.values()), "Incomplete matrix read!"
        if not taxlabels:
            assert not format.transpose
            taxlabels = {i + 1: key for i, key in enumerate(res)}

        def apply_to_state(func, state, *args, **kw):
            # We have to do a bit of mapping and renaming, which always needs to deal with the
            # three different types of state.
            if isinstance(state, str):
                return func(state, *args, **kw)
            if isinstance(state, tuple):
                return tuple(func(s, *args, **kw) for s in state)
            if isinstance(state, set):
                return set(func(s, *args, **kw) for s in state)
            raise ValueError(state)  # pragma: no cover

        lax_symbols = not format.explicit_symbols and format.datatype in {None, 'STANDARD'} \
            and not (self.nexus and self.nexus.cfg.strict)

        def replace_symbol(s, i, r):
            if (format.respectcase and s == format.missing) or \
                    (not format.respectcase and (s.upper() == format.missing.upper())):
                return None
            if format.gap:
                if (format.respectcase and s == format.gap) or \
                        (not format.respectcase and (s.upper() == format.gap.upper())):
                    return GAP
            if format.matchchar:  # match entries from first row!
                if (format.respectcase and s == format.matchchar) or \
                        (not format.respectcase and (s.upper() == format.matchchar.upper())):
                    assert r
                    return r[i]
            if s not in format.symbols:
                s = s.lower() if s.isupper() else s.upper()

            if not lax_symbols:
                assert s in format.symbols, '{} {}'.format(s, format.symbols)
            return s

        def resolve_symbols(s, i, r):
            def resolve(c, i, r):
                c = format.equate.get(c.upper(), c)  # May result in ambiguous or multiple states!
                return apply_to_state(replace_symbol, c, i, r)
            return apply_to_state(resolve, s, i, r)

        firstrow = None
        for i, l in enumerate(res):
            res[l] = [resolve_symbols(s, i, firstrow) for i, s in enumerate(res[l])]
            if i == 0:
                # We need the fully resolved entries of the first row around to resolve MATCHCHARs.
                firstrow = res[l]

        # Create the final result, an OrderedDict mapping taxa labels (or numbers) to OrderedDicts
        # mapping character labels or numbers to state symbols or labels.
        matrix = collections.OrderedDict()
        if not format.transpose:
            tlabels = {str(k) for k in res.keys()}
            valid_taxa = {str(k) for k in taxlabels}.union(taxlabels.values())
            if not tlabels.issubset(valid_taxa):
                if self.nexus and self.nexus.cfg.strict:  # pragma: no cover
                    raise ValueError('Found undeclared taxa in characters matrix')
                else:
                    warnings.warn('Dropping undeclared taxa from characters matrix.')

        for tnum, tlabel in sorted(taxlabels.items()):
            if format.transpose:
                # We have to pick the tnum column in each list in res.
                matrix[tlabel] = collections.OrderedDict()
                for cnum, clabel in sorted(charlabels.items()):
                    entries = res[clabel] if clabel in res else res[cnum]
                    matrix[tlabel][clabel] = entries[tnum - 1]
            else:
                key = tlabel if tlabel in res else (tnum if tnum in res else str(tnum))
                if key in res:
                    # Non-transposed matrices may not have data for each taxon!
                    entries = res[key]
                    matrix[tlabel] = collections.OrderedDict(
                        [(charlabels[i], s) for i, s in enumerate(entries, start=1)])
        if labeled_states:
            for entries in matrix.values():
                for char in entries:
                    if char in statelabels:
                        if entries[char] not in {None, GAP}:
                            entries[char] = apply_to_state(
                                lambda s: statelabels[char].get(s) or s, entries[char])
        return matrix

    def get_taxlabels(self, format):
        if self.TAXLABELS:
            taxlabels = self.TAXLABELS.labels
            ntax = self.DIMENSIONS.ntax
        elif 'TAXA' in self.linked_blocks:
            taxlabels = self.linked_blocks['TAXA'].TAXLABELS.labels
            ntax = self.linked_blocks['TAXA'].DIMENSIONS.ntax
        elif self.nexus.TAXA:
            taxlabels = self.nexus.TAXA.TAXLABELS.labels
            ntax = self.nexus.TAXA.DIMENSIONS.ntax
        else:
            ntax, taxlabels = None, {}

        if format.interleave and format.labels is False and not format.transpose:
            # If the matrix has no row labels and is not transposed, we need the number of taxa to
            # compute the size of the interleaved blocks.
            assert ntax
            taxlabels = taxlabels or {i + 1: str(i + 1) for i in range(ntax)}
        return ntax, taxlabels

    def get_charstatelabels(self, nchar=None, format=None):
        nchar = nchar or self.DIMENSIONS.nchar
        charlabels = {i + 1: str(i + 1) for i in range(nchar)}
        statelabels = {}

        if self.CHARSTATELABELS:
            charlabels = {
                int(c.number): c.name or str(c.number) for c in self.CHARSTATELABELS.characters}
            statelabels = {
                c.number: c.states for c in self.CHARSTATELABELS.characters}
        elif self.CHARLABELS:
            charlabels = {
                int(c.number): c.name or str(c.number) for c in self.CHARLABELS.characters}
        if self.STATELABELS:
            statelabels = {c.number: c.states for c in self.STATELABELS.characters}

        format = format or self.FORMAT or Format(None)
        if statelabels:
            statelabels = {charlabels[cnum]: states for cnum, states in statelabels.items()}
            for clabel in statelabels:
                states = statelabels[clabel]
                labeled = collections.OrderedDict()
                if format:
                    for i, symbol in enumerate(format.symbols):
                        if i < len(states) and states[i] != '_':
                            labeled[symbol] = states[i]
                statelabels[clabel] = labeled
        else:
            statelabels = {}
        assert len(charlabels) == nchar
        return charlabels, statelabels

    def validate(self, log=None):
        res = super().validate(log)
        if 'TAXLABELS' in self.commands and not self.DIMENSIONS.newtaxa:
            return log_or_raise(
                'TAXLABELS may only be defined in {} block if NEWTAXA is specified.'.format(
                    self.name),
                log=log)
        return res

    @classmethod
    def from_data(cls,
                  matrix: StateMatrix,
                  taxlabels: bool = False,
                  statelabels: typing.Optional[typing.Dict[str, typing.Dict[str, str]]] = None,
                  datatype: str = 'STANDARD',
                  missing: str = '?',
                  gap: str = '-',
                  TITLE: typing.Optional[str] = None,
                  ID: typing.Optional[str] = None,
                  LINK: typing.Optional[typing.Union[str, typing.Tuple[str, str]]] = None,
                  **kw) -> 'Characters':
        """
        Instantiate a CHARACTERS or DATA block from a metrix.

        This functionality can be used to normalize the NEXUS formatting of CHARACTERS matrices:

        .. code-block:: python

            >>> nex = Nexus('''#NEXUS
            ... BEGIN TAXA;
            ... DIMENSIONS NTAX=3;
            ... TAXLABELS t1 t2 t3;
            ... END;
            ... BEGIN CHARACTERS;
            ... DIMENSIONS NCHAR=3;
            ... FORMAT TRANSPOSE NOLABELS;
            ... MATRIX 100 010 001;
            ... END;''')
            >>> matrix = nex.CHARACTERS.get_matrix()
            >>> nex.replace_block(nex.CHARACTERS, Characters.from_data(matrix))
            >>> print(nex)
            #NEXUS
            BEGIN TAXA;
            DIMENSIONS NTAX=3;
            TAXLABELS t1 t2 t3;
            END;
            BEGIN CHARACTERS;
            DIMENSIONS NCHAR=3;
            FORMAT DATATYPE=STANDARD MISSING=? GAP=- SYMBOLS="01";
            MATRIX
            t1 100
            t2 010
            t3 001
            ;
            END;

        :param matrix: A matrix as returned by :meth:`Characters.get_matrix()`, with unlabeled \
        states. I.e. `None` is used to mark missing values, and `GAP` to mark gapped values. These \
        special states will be converted to the symbols passed as `missing` and `gap` upon writing.
        :param taxlabels: If `True`, include a TAXLABELS command rather than relying on a TAXA \
        block being present.
        :param datatype:
        :param missing:
        :param gap:
        :param kw:
        """
        if datatype != 'STANDARD':  # pragma: no cover
            raise NotImplementedError('Only DATATYPE=STANDARD is supported for writing CHARACTERS')
        nexus = kw.pop('nexus', None)

        symbols, rows, charlabels, maxlen, tlabels = set(), [], None, 0, {}
        for taxon in matrix:  # We compute maximum taxon label length for pretty printing.
            tlabels[taxon] = Word(taxon).as_nexus_string()
            maxlen = max([maxlen, len(tlabels[taxon])])

        symbol = lambda c: missing if c is None else (gap if c == GAP else c)  # noqa: E731

        for taxon, entries in matrix.items():
            if not charlabels:
                charlabels = collections.OrderedDict(
                    [(str(i + 1), c) for i, c in enumerate(entries)])
            row = []
            for entry in entries.values():
                if entry:
                    symbols |= set(entry)
                if isinstance(entry, tuple):  # polymorphism -> ()
                    row.append('({})'.format(''.join(symbol(c) for c in entry)))
                elif isinstance(entry, set):  # uncertainty -> {}
                    row.append('{{{}}}'.format(''.join(sorted(symbol(c) for c in entry))))
                else:
                    row.append(symbol(entry))
            rows.append('\n{} {}'.format(tlabels[taxon].ljust(maxlen), ''.join(row)))

        symbols = ''.join(sorted([s for s in symbols if s not in [None, GAP]]))
        if missing in symbols or (gap in symbols):
            raise ValueError('MISSING or GAP markers must be distinct from "{}"'.format(symbols))
        respectcase = any(c.isupper() for c in symbols) and any(c.islower() for c in symbols)

        dimensions = 'NCHAR={}'.format(len(list(matrix.values())[0]))
        if taxlabels:
            dimensions = 'NEWTAXA NTAX={} {}'.format(len(tlabels), dimensions)
        cmds = [
            ('DIMENSIONS', dimensions),
            ('FORMAT', 'DATATYPE=STANDARD {}MISSING={} GAP={} SYMBOLS="{}"'.format(
                'RESPECTCASE ' if respectcase else '', missing, gap, symbols)),
        ]
        statelabels = statelabels or {}
        if any(k != v for k, v in charlabels.items()):
            cmds.append((
                'CHARSTATELABELS',
                ', '.join('\n    {} {}{}'.format(
                    n,
                    Word(l).as_nexus_string(),
                    '/' + ' '.join(Word(ll).as_nexus_string() for ll in statelabels[l].values())
                    if statelabels.get(l) else '',
                ) for n, l in charlabels.items())))
        if taxlabels:
            cmds.append(('TAXLABELS', ' '.join(tlabels.values())))
        cmds.append(('MATRIX', ''.join(rows) + '\n'))
        return cls.from_commands(cmds, nexus=nexus, TITLE=TITLE, ID=ID, LINK=LINK)


class Options(Payload):
    """
    The GAPMODE subcommand of the OPTIONS command of the ASSUMPTIONS block was originally
    housed in an OPTIONS command in the DATA block.

    :ivar typing.Optional[str] gapmode: `missing` or `newstate`.
    """
    def __init__(self, tokens, nexus=None):
        super().__init__(tokens, nexus=nexus)
        self.gapmode = None

        words = iter_words_and_punctuation(self._tokens, nexus=nexus)

        while 1:
            try:
                word = next(words)
                if isinstance(word, str) and word.upper() == 'GAPMODE':
                    self.gapmode = word_after_equals(words).lower()
            except StopIteration:
                break
        assert self.gapmode in {None, 'missing', 'newstate'}


class Data(Characters):
    """
    This block is equivalent to a CHARACTERS block in which the NEWTAXA subcommand is included in
    the DIMENSIONS command. That is, the DATA block is a CHARACTERS block that includes not only
    the definition of characters but also the definition of taxa.

    .. note::

        The GAPMODE subcommand of the OPTIONS command of the ASSUMPTIONS block was originally
        housed in an :class:`OPTIONS <Options>` command in the DATA block.
    """
    __commands__ = [
        Dimensions, Format, Options,
        Eliminate, Taxlabels, Charstatelabels, Charlabels, Statelabels, Matrix]



"""
Manipulate the CHARACTERS (or DATA) block of a NEXUS file.

Note: Only one option can be chosen at a time.
"""
import collections
from enum import Enum

from commonnexus import Nexus
from commonnexus.blocks.characters import Characters, GAP
from commonnexus.tools.matrix import CharacterMatrix
from commonnexus.tools.combine import combine
from commonnexus.cli_util import add_nexus, add_flag, validate_operations, list_of_ranges


class Ops(Enum):
    binarise = 1
    multistatise = 2
    convert = 3
    drop = 4
    drop_numbered = 5
    describe = 6


def register(parser):
    add_nexus(parser)
    add_flag(
        parser,
        Ops.binarise.name,
        help="Recode a matrix such that it only contains binary characters.")
    parser.add_argument(
        '--' + Ops.multistatise.name,
        metavar="GROUPKEY",
        default=None,
        help="Recode a matrix such that it only contains one multistate characters for each "
             "group of characters as determined by GROUPKEY, a Python lambda function accepting "
             "character label and returning a key (or a string to group all characters into one "
             "multistate character labeled with this string).")
    parser.add_argument(
        '--' + Ops.convert.name,
        choices=['fasta', 'phylip'],
        help="Convert a matrix to another sequence format.")
    parser.add_argument(
        '--' + Ops.drop.name,
        help="Drop specified characters from a matrix.",
        choices='constant polymorphic uncertain missing gapped'.split())
    parser.add_argument(
        '--' + Ops.drop_numbered.name.replace('_', '-'),
        type=list_of_ranges,
        help="Drop characters specified by (ranges of) numbers from a matrix.")
    parser.add_argument(
        '--' + Ops.describe.name,
        help="",
        choices=['binary-setsize', 'binary-unique', 'binary-constant', 'states-distribution'])


def run(args):
    validate_operations(Ops, args)
    new = None
    if args.binarise:
        if args.nexus.characters.FORMAT:  # pragma: no cover
            assert args.nexus.characters.FORMAT.datatype is None or \
                args.nexus.characters.FORMAT.datatype == 'STANDARD'
        _, statelabels = args.nexus.characters.get_charstatelabels()
        new = CharacterMatrix.binarised(args.nexus.characters.get_matrix(), statelabels=statelabels)
    elif args.multistatise:
        if args.multistatise.startswith('lambda'):
            groupkey = eval(args.multistatise)
        else:
            groupkey = lambda c: args.multistatise  # noqa: E731
        charpartitions = collections.defaultdict(list)
        matrix = CharacterMatrix(args.nexus.characters.get_matrix())
        for char in matrix.characters:
            charpartitions[groupkey(char)].append(char)
        matrices = [
            (key, CharacterMatrix.from_characters(matrix, drop_chars=chars, inverse=True))
            for key, chars in charpartitions.items()]
        if len(matrices) == 1:
            new = CharacterMatrix.multistatised(matrices[0][1], multicharlabel=matrices[0][0])
            args.nexus.replace_block(args.nexus.characters, Characters.from_data(new))
            new = args.nexus
        else:
            ms = []
            for key, m in matrices:
                ms.append(CharacterMatrix.multistatised(m, multicharlabel=key))
            new = combine(*[Nexus.from_blocks(Characters.from_data(m)) for m in ms])
        print(new)
        return
    elif args.drop:
        #
        # FIXME: drop unique - for binary matrices
        #
        kw = {'drop_{}'.format(args.drop): True}
        new = CharacterMatrix.from_characters(args.nexus.characters.get_matrix(), **kw)
    elif args.drop_numbered:
        m = CharacterMatrix(args.nexus.characters.get_matrix())
        kw = {'drop_chars': [
            c for i, c in enumerate(m.characters, start=1) if i in args.drop_numbered]}
        new = CharacterMatrix.from_characters(m, **kw)
    elif args.convert:
        m = CharacterMatrix(args.nexus.characters.get_matrix())
        print(getattr(m, 'to_' + args.convert)())
        return
    elif args.describe:
        m = CharacterMatrix(args.nexus.characters.get_matrix())
        chars = m.characters
        chars = collections.OrderedDict([(chars[i], col) for i, col in enumerate(m.iter_columns())])

        if args.describe == 'binary-setsize':
            assert m.is_binary
            args.log.info('Character set sizes (for binary matrix):')
            for k, v in chars.items():
                print('{} {}'.format(k, sum(1 for vv in v if vv and '1' in vv)))
        elif args.describe == 'binary-unique':
            assert m.is_binary
            args.log.info('Unique character sets (for binary matrix):')
            for k, v in chars.items():
                if sum(1 for vv in v if vv and '1' in vv) == 1:
                    print(k)
        elif args.describe == 'binary-constant':
            assert m.is_binary
            args.log.info('Constant characters (for binary matrix):')
            for k, v in chars.items():
                if len(set(v)) == 1:
                    print(k)
        elif args.describe == 'states-distribution':
            args.log.info('State distribution\nchar\tstate\tcount')
            for char, states in chars.items():
                states = collections.Counter(
                    [frozenset(v) if isinstance(v, set) else v for v in states])
                for k, v in states.most_common():
                    print('{}\t{}\t{}'.format(
                        char, '?' if k is None else ('-' if k == GAP else k), v))
        return

    if new:
        args.nexus.replace_block(args.nexus.characters, Characters.from_data(new))
        print(args.nexus)
        return
    matrix = CharacterMatrix(args.nexus.characters.get_matrix())
    args.log.info('{} block with {} taxa and {} characters.'.format(
        args.nexus.characters.name, len(matrix.taxa), len(matrix.characters)))
